# GUI.pm
# contains a Graphical User Interface for the aVTCtest evaluation system
# Copyright (C) 2003-2004 Michel Messerschmidt
#
package Avtc::GUI;
use strict;
use Exporter;
use File::Basename;
use Tk;
use Tk::Tiler;
use Tk::LabFrame;
use Tk::Pane;
use Tk::Radiobutton;
use Tk::Message;
use Tk::DialogBox;
use Tk::Text;
use Tk::ROText;
use Tk::Balloon;
use Tk::BrowseEntry;

@Avtc::GUI::ISA = qw(Exporter);
@Avtc::GUI::EXPORT = qw(startGUI);

my $guiVersion = "v0.41, 2004-07-09 written by Michel Messerschmidt";

# flag for Administrator mode (may be change via menu entry)
# Some commands are only available in AdministratorMode
my $adminMode = 0;
# These hashes hold the currently selected options
# Note that nothing is selected per default (empty hashes)
# This is different from the commandline defaults, where
# ALL systems/testbeds/products are selected per default
my %selectOs = ();
my %selectTb = ();
my %selectProd = ();
my %selectPack = ();
#popup help widget
my $popHelp;
# frame with selected test display
my $titleLabel;
# frame with selection boxes
my $selFrame;
# list of active widgeds:
# each key is a parameter type (from %::cmds)
# each value is a list of widgets that should be activated for this type
my %activeWidgets = ();
foreach my $t ("o", "t", "p", "a", "s") {
    @{$activeWidgets{$t}} = ();
}
# list of commands and their widgets:
my %cmdWidgets = ();
# widget that contains the 'execute' button
my $execWidget;
# widget containing all command buttons
# ('updateCommands' may need to access this)
my $cmdTiler = ();
# some often used settings
my $fontNorm;
my $fontBold;
my $fontNormP;
my $fontPBold;
my $fontPop;
my $colorNorm = 'black';
my $colorAlert = 'red';
my (@f) =  (qw/-relief flat -borderwidth 1/);
my (@fr) = (qw/-relief ridge -borderwidth 1/);
my (@fs) = (qw/-relief solid -borderwidth 1/);
my (@p) =  (qw/-anchor nw -side top -pady 0 -expand yes -fill both/);
my (@pl) = (qw/-anchor n -side top -expand yes -fill x/);
my (@c) =  (qw/-borderwidth 1 -padx 0 -pady 0 -relief raised/);
my (@cf) = (qw/-borderwidth 1 -padx 0 -pady 0 -relief flat/);


sub startGUI {
    # create all window elements
    $::mainWin = MainWindow->new(-title => 'aVTC Test Evaluation');
    # define some output windows elements (that must be accessed from menubar)
    my $outScrolled;
    my $linewrap = "word";
    # get default foregrond color
    $colorNorm = $::mainWin->cget('-foreground');
    # create fonts (can be selected from menubar)
    my $fontsize = 11;
    my %font = ();
    foreach my $s (9, 10, 11, 12, 14, 16, 18, 20, 24) {
        $font{$s} = $::mainWin->fontCreate("text".$s, -family => "courier",
                                                 -size => $s);
        $font{"p".$s} = $::mainWin->fontCreate("prop".$s, -family => "helvetica",
                                                 -size => $s);
        $font{"p".$s."bold"} = $::mainWin->fontCreate("prop".$s."b",
                                             -family => "helvetica",
                                             -size => $s,
                                             -weight => "bold");
        $font{$s."bold"} = $::mainWin->fontCreate("text".$s."b", -family => "courier",
                                                       -size => $s,
                                                       -weight => "bold");
    }
    $fontNorm = $font{$fontsize};
    $fontBold = $font{$fontsize."bold"};
    $fontNormP = $font{"p".$fontsize};
    $fontPBold = $font{"p".$fontsize."bold"};
    $fontPop = $font{"p10"};
    # create popup help widget for all changing widgets (selection frame contents)
    $popHelp = $::mainWin->Balloon(-state => 'balloon',
                                -balloonposition => 'mouse',
                                -font => $fontPop);
    # create popup help widget for all non-changing widgets
    my $popHelpConst = $::mainWin->Balloon(-state => 'balloon',
                                -balloonposition => 'mouse',
                                -font => $font{"p10"});

    # create menu
    my $menu = $::mainWin->Menu(-borderwidth => 1);
    $::mainWin->configure(-menu => $menu);

    my $menuGen = $menu->Menu(-tearoff => 0);
    $menu->add("cascade", -label => "General",
                          -underline => 0,
                          -menu => $menuGen);
    my $menuGenTest = $menuGen->Menu(-tearoff => 0);
    $menuGen->add("cascade", -label => "Select test",
                             -underline => 0,
                             -menu => $menuGenTest);
    foreach my $i (sort keys %::configAllTests) {
        $menuGenTest->add("radiobutton", -variable => \$::configTestName,
                                          -underline => 1,
                                          -value => $i,
                                          -label => " $i",
                                          -command => sub { updateTest() });
    }
    $menuGen->add("checkbutton", -label => " Administrator Mode",
                                 -underline => 1,
                                 -variable => \$adminMode,
                                 -onvalue => 1, offvalue => 0,
                                 -command => sub { updateCommands();
                                                   if ($adminMode) {
                                                       ::log("Administrator Mode enabled.\n");
                                                   } else {
                                                       ::log("Administrator Mode disabled.\n");
                                                   }
                                                 }
                 );
    $menuGen->add("separator");
    $menuGen->add("command", -label => "Exit",
                             -underline => 1,
                             -command => sub { ::exitPrg(0) });

    my $menuOpt = $menu->Menu(-tearoff => 0);
    $menu->add("cascade", -label => "Options",
                          -underline => 0,
                          -menu => $menuOpt);
    my $menuOptDebug = $menuOpt->Menu(-tearoff => 0);
    $menuOpt->add("cascade", -label => " Debug Level",
                             -underline => 1,
                             -menu => $menuOptDebug);
    $menuOptDebug->add("radiobutton", -variable => \$::debug,
                                      -underline => 1,
                                      -value => 0,
                                      -label => " 0 (off)");
    $menuOptDebug->add("radiobutton", -variable => \$::debug,
                                      -underline => 1,
                                      -value => 1,
                                      -label => " 1 (verbose)");
    $menuOptDebug->add("radiobutton", -variable => \$::debug,
                                      -underline => 1,
                                      -value => 2,
                                      -label => " 2 (key debugging)");
    $menuOptDebug->add("radiobutton", -variable => \$::debug,
                                      -underline => 1,
                                      -value => 3,
                                      -label => " 3 (all)");
    $menuOptDebug->add("radiobutton", -variable => \$::debug,
                                      -underline => 1,
                                      -value => 4,
                                      -label => " 4 (not used)");
    $menuOptDebug->add("radiobutton", -variable => \$::debug,
                                      -underline => 1,
                                      -value => 5,
                                      -label => " 5 (internal debugging)");
    my $menuOptFont = $menuOpt->Menu(-tearoff => 0);
    $menuOpt->add("cascade", -label => " Font size",
                             -underline => 1,
                             -menu => $menuOptFont);
    foreach my $s (10, 11, 12, 14, 16, 18, 20, 24) {
        $menuOptFont->add("radiobutton", -variable => \$fontsize,
                                         -underline => 2,
                                         -value => $s,
                                         -label => " $s",
                                         -command => sub {
                                             $outScrolled->configure(
                                                 -font => $font{$s});
                                         }
        );
    }
    $menuOpt->add("checkbutton", -label => " Line wrap",
                                 -underline => 1,
                                 -variable => \$linewrap,
                                 -onvalue => "word", offvalue => "none",
                                 -command => sub { $outScrolled->configure(-wrap => $linewrap);
                                 }
    );


    my $menuHelp = $menu->Menu(-tearoff => 0);
    $menu->add("cascade", -label => "Help",
                          -underline => 0,
                          -menu => $menuHelp);
    $menuHelp->add("command", -label => "Command Help",
                              -underline => 8,
                              -command => sub { showDocWindow("Help", 
                                                    join("", @::usage));
                                              });
    $menuHelp->add("separator");
    $menuHelp->add("command", -label => "Documentation Overview",
                              -underline => 0,
                              -command => sub { showDoc("index.txt") });
    $menuHelp->add("command", -label => "Locations (Was ist wo ?)",
                              -underline => 11,
                              -command => sub { showDoc("wasistwo.txt") });
    $menuHelp->add("command", -label => "Introduction to Test Evaluation",
                              -underline => 0,
                              -command => sub { showDoc("avtctest.txt") });
    $menuHelp->add("command", -label => "Control File Help",
                              -underline => 0,
                              -command => sub { showDoc("control.txt") });
    $menuHelp->add("command", -label => "Testbeds [german]",
                              -underline => 0,
                              -command => sub { showDoc("testbed.txt") });
    $menuHelp->add("command", -label => "Evaluation Files",
                              -underline => 11,
                              -command => sub { showDoc("files.txt") });
    $menuHelp->add("command", -label => "Evaluation Lists",
                              -underline => 11,
                              -command => sub { showDoc("lists.txt") });
    $menuHelp->add("command", -label => "Evaluation Results",
                              -underline => 11,
                              -command => sub { showDoc("result.txt") });
    $menuHelp->add("command", -label => "Status Codes",
                              -underline => 0,
                              -command => sub { showDoc("status.txt") });
    $menuHelp->add("command", -label => "Scanner Keys",
                              -underline => 8,
                              -command => sub { showDoc("scanner-key.txt") });
    $menuHelp->add("command", -label => "New Test setup [german]",
                              -underline => 0,
                              -command => sub { showDoc("new_test.txt") });
    $menuHelp->add("command", -label => "Packed Testbed Evaluation [german]",
                              -underline => 1,
                              -command => sub { showDoc("packed-eval.txt") });
    $menuHelp->add("command", -label => "Packed Testbed Creation [german]",
                              -underline => 0,
                              -command => sub { showDoc("create_packed.txt")});
    $menuHelp->add("command", -label => "Internal data structures",
                              -underline => 0,
                              -command => sub {showDoc("data-structures.txt")});
    $menuHelp->add("separator");
    $menuHelp->add("command", -label => "About",
                              -underline => 1,
                              -command => sub { showAbout() });

    # basic frames:
    # frame for text output
    my $outFrame = $::mainWin->LabFrame(@f, -label => "Log Output",
                                   -labelside => "acrosstop");
    # frame with selection boxes
    $selFrame = $::mainWin->Frame(@f);
    # frame with selected test display
    my $testFrame = $::mainWin->LabFrame(@f, -label => "Selected test",
                                -labelside => "acrosstop");
    # frame with command buttons
    my $cmdFrame = $::mainWin->LabFrame(@f, -label => "Available commands",
                                -labelside => "acrosstop");
    # frame with avtctest command buttons
    my $execFrame = $::mainWin->Frame(-relief => "flat", -borderwidth => 5);

    # set layout
    $outFrame->form(-top => '%0', -left => $selFrame,
                    -right => '%100', -bottom => '%100');
    $testFrame->form(-top => '%0', -left => '%0',
                    -right => $outFrame, -bottom => 'none');
    $cmdFrame->form(-top => $testFrame, -left => '%0',
                    -right => $outFrame, -bottom => 'none');
    $selFrame->form(-top => $cmdFrame, -left => '%0',
                    -right => 'none', -bottom => $execFrame);
    $execFrame->form(-top => 'none', -left => '%0',
                    -right => $outFrame, -bottom => '%100');

    $titleLabel = $testFrame->Label(-text => '-- no test selected --',
                                    -anchor => "center",
                                    -justify => "center");
    $titleLabel->form(-top => '%0', -left => '%0',
                      -right => '%100', -bottom => '%100');

    # output frame contents
    $outScrolled = $outFrame->Scrolled('ROText', -scrollbars => "se",
                                    -wrap           => $linewrap,
                                    -width          => 60,
                                    -height         => 30,
                                    -font           => $fontNorm,
                                    -setgrid        => 1,
    )->pack(-side => 'left', -expand => "yes", -fill => "both");
    $::outMsg = $outScrolled->Subwidget('scrolled');
    $::outMsg->insert("end", $::prog." ".$::version."\n");

    # exec frame contents
    my $exec2 = $execFrame->Button(-text => 'Exit',
                                   -command => sub { ::exitPrg(0) })
                          ->pack(-side => 'left', -padx => 10, -pady => 2);
    $popHelpConst->attach($exec2, -msg => 'Terminate the avtctest program');
    $execWidget = $execFrame->Button(-text => 'Make it so',
                                   -padx => 50,
                                   -command => sub { doCmd() })
                          ->pack(-side => 'left', -padx => 10, -pady => 2);
    $popHelpConst->attach($execWidget, -msg =>
        'Execute the currently selected avtctest command with currently '.
        'selected parameters');
    my $exec3 = $execFrame->Button(-text => 'Clear Log',
                                   -command => sub { $::outMsg->delete("1.0", "end" ) })
                          ->pack(-side => 'right', -padx => 10, -pady => 2);
    $popHelpConst->attach($exec3, -msg => 'Clear the log window (but not the logfile)');

    # command frame contents
    $cmdTiler = $cmdFrame->Scrolled('Tiler', -columns => 2, -rows => 5,
                                       -scrollbars => "oe");
    my @cl = qw/-borderwidth 1 -padx 0 -pady 0 -relief flat -anchor nw -justify left/;

    foreach my $cmd (sort keys %::cmds) {
        if ($::cmds{$cmd}->{"notGUI"}) { next }
        my $label = $::cmds{$cmd}->{"label"};
        my $desc = $::cmds{$cmd}->{"desc"};
        my $b = $cmdTiler->Radiobutton(@cl,
                                       -text => $label,
                                       -value => $cmd,
                                       -variable => \$::command,
                                       -command => sub { updateActivation() }
                                       );
        $popHelpConst->attach($b, -msg => \$desc);
        $cmdTiler->Manage($b);
        # add to hash to access widgets via command name
        $cmdWidgets{$cmd} = $b;
    }
    $cmdTiler->form(-top => '%0', -left => '%0',
                    -right => '%100', -bottom => '%100');


    # set default command
    $::command = "show status";
    # scroll to default command
    $cmdTiler->yview("scroll", 10, "units");

    # read test configuration and setup selection frame contents
    updateTest();

    # set flag for other functions that gui elements are available
    $::gui = 1;
    # start the GUI
    MainLoop;
}


# execute the selected commands with the selected options
sub doCmd {
    # handle os selection
    %::osList = ();
    foreach my $i (keys %selectOs) {
       if ($selectOs{$i}) {
           $::osList{$i} = $::osListAll{$i};
       }
    }
    # handle testbed selection
    %::lists = ();
    foreach my $i (keys %selectTb) {
       if ($selectTb{$i}) {
           $::lists{$i} = $::listsAll{$i};
       }
    }
    # handle product selection
    %::prodList = ();
    foreach my $i(keys %selectProd) {
       if ($selectProd{$i}) {
           $::prodList{$i} = $::prodListAll{$i};
       }
    }
    # handle packer selection
    %::packList = ();
    foreach my $i(keys %selectPack) {
       if ($selectPack{$i}) {
           $::packList{$i} = $::packListAll{$i};
       }
    }

    # output all options on higher debuglevels
    if ($::debug > 1) {
        ::log("Selected os:", 1);
        foreach (keys %::osList) {
            ::log(" $_");
        }
        ::log("\n");
        ::log("Selected testbeds:", 1);
        foreach (keys %::lists) {
            ::log(" $_");
        }
        ::log("\n");
        ::log("Selected products:", 1);
        foreach (keys %::prodList) {
            ::log(" $_");
        }
        ::log("\n");
        ::log("Selected packer:", 1);
        foreach (keys %::packList) {
            ::log(" $_");
        }
        ::log("\n");
    }


    ### execute related function
    if ($::cmds{$::command}->{"ask"}) {
        my $cancel = confirmDialog();
        if ($cancel) {
            ::log("Execution of command '$::command' canceled\n", 0);
        } else {
            # deactivate all command and selection widgets while executing a command
            deactivateWidgets();
            ::execCmd();
        }
    } else {
        deactivateWidgets();
        ::execCmd();
    }
    # reactivate widgets
    updateCommands();
    updateActivation();
    $execWidget->configure(-state => "normal");
}


# Ask for confimation before executing a command
sub confirmDialog {
    # default layout for inner labels
    my (@l) =  (qw/-anchor nw -justify left/);
    my (@lc) =  (qw/-anchor center -justify center/);
    my $font;
    my $color;
    # create dialog widget
    my $dlg = $::mainWin->DialogBox(-title => "Confirm execution",
                               -buttons => ["Ok", "Cancel"],
                               -default_button => "Cancel");
    my $cmdL1 = $dlg->add("Label", @lc,
              -text => "Do you really want to execute the command :\n")
        ->form(-top => '%0', -bottom => 'none',
                    -left => '%0', -right => '%100');
    my $cmdL2 = $dlg->add("Label", -anchor => 'center', -justify => 'center', 
                -padx => 10, 
                -relief => 'solid', -font => $fontBold, -text => $::command)
        ->form(-top => $cmdL1, -bottom => 'none',
                    -left => '%0', -right => '%100');
    my $cmdL3 = $dlg->add("Label", -anchor => 'nw', -justify => 'center', 
                          -padx => 0, 
                          -text => "(". $::cmds{$::command}{"desc"} .")\n")
        ->form(-top => $cmdL2, -bottom => 'none',
                    -left => '%0', -right => 'none');
    # get command parameters
    my $osL1 = $dlg->add("Label", -anchor => 'nw', -justify => 'right', 
                         -pady => 0, -padx => 10, -text => "For systems: ")
        ->form(-top => $cmdL3, -bottom => 'none',
                    -left => '%0', -right => 'none');
    my $osString = "";
    my $allOs = 1;
    foreach my $i (keys %::osListAll) {
        if ($selectOs{$i} == 0) {
            $allOs = 0;
            last;
        }
    }
    if ($allOs == 1) {
        $osString = "ALL";
        $font = $fontBold;
        $color = $colorAlert;
    } else {
        $font = $fontNorm;
        $color = $colorNorm;
        foreach my $i (keys %::osList) {
            $osString .= " " . $i . ",";
        }
        chop($osString);
    }
    my $osL2 = $dlg->add("Label", -anchor => 'nw', -justify => 'left', 
                         -pady => 0, -padx => 10, -text => $osString,
                         -foreground => $color, -font => $font)
        ->form(-top => $cmdL3, -bottom => 'none',
                    -left => $osL1, -right => '%100');

    my $tbL1 = $dlg->add("Label", -anchor => 'nw', -justify => 'right', 
                         -pady => 0, -padx => 10, -text => "For testbeds: ")
        ->form(-top => $osL1, -bottom => 'none',
                    -left => '%0', -right => 'none');
    my $tbString = "";
    my $allTb = 1;
    foreach my $i (keys %::listsAll) {
        if ($selectTb{$i} == 0) {
            $allTb = 0;
            last;
        }
    }
    if ($allTb == 1) {
        $tbString = "ALL";
        $font = $fontBold;
        $color = $colorAlert;
    } else {
        $font = $fontNorm;
        $color = $colorNorm;
        foreach my $i (keys %::lists) {
            $tbString .= " " . $i . ",";
        }
        chop($tbString);
    }
    my $tbL2 = $dlg->add("Label", -anchor => 'nw', -justify => 'left', 
                         -pady => 0, -padx => 10, -text => $tbString,
                         -foreground => $color, -font => $font)
        ->form(-top => $osL2, -bottom => 'none',
                    -left => $tbL1, -right => '%100');

    my $prodL1 = $dlg->add("Label", -anchor => 'nw', -justify => 'right', 
                         -pady => 0, -padx => 10, -text => "For products: ")
        ->form(-top => $tbL1, -bottom => 'none',
                    -left => '%0', -right => 'none');
    my $prodString = "";
    my $allProd = 1;
    foreach my $i (keys %::prodListAll) {
        if ($selectProd{$i} == 0) {
            $allProd = 0;
            last;
        }
    }
    if ($allProd == 1) {
        $prodString = "ALL";
        $font = $fontBold;
        $color = $colorAlert;
    } else {
        $font = $fontNorm;
        $color = $colorNorm;
        foreach my $i (keys %::prodList) {
            $prodString .= " " . $i . ",";
        }
        chop($prodString);
    }
    my $prodL2 = $dlg->add("Label", -anchor => 'nw', -justify => 'left', 
                         -pady => 0, -padx => 10, -text => $prodString,
                         -foreground => $color, -font => $font)
        ->form(-top => $tbL2, -bottom => 'none',
                    -left => $prodL1, -right => '%100');

    my $packL1 = $dlg->add("Label", -anchor => 'nw', -justify => 'right', 
                         -pady => 0, -padx => 10, -text => "For packer: ")
        ->form(-top => $prodL1, -bottom => 'none',
                    -left => '%0', -right => 'none');
    my $packString = "";
    my $allPack = 1;
    foreach my $i (keys %::packListAll) {
        if ($selectPack{$i} == 0) {
            $allOs = 0;
            last;
        }
    }
    if ($allPack == 1) {
        $packString = "ALL";
        $font = $fontBold;
        $color = $colorAlert;
    } else {
        $font = $fontNorm;
        $color = $colorNorm;
        foreach my $i (keys %::packList) {
            $packString .= " " . $i . ",";
        }
        chop($packString);
    }
    my $packL2 = $dlg->add("Label", -anchor => 'nw', -justify => 'left', 
                         -pady => 0, -padx => 10, -text => $packString,
                         -foreground => $color, -font => $font)
        ->form(-top => $prodL2, -bottom => 'none',
                    -left => $packL1, -right => '%100');

    if ($::command eq "reset" 
        or $::command eq "reset ok" or $::command eq "reset postscan"
        or $::command eq "set ok" or $::command eq "set postscan") {
        my $statL1 = $dlg->add("Label", -anchor => 'nw', -justify => 'right', 
                             -pady => 0, -padx => 10, -text => "New status: ")
            ->form(-top => $packL1, -bottom => 'none',
                        -left => '%0', -right => 'none');
        my $statString = sprintf("%10s (%s)", $::resetStatus,
                                 $::statusCodes{$::resetStatus});
        my $statL2 = $dlg->add("Label", -anchor => 'nw', -justify => 'left', 
                             -pady => 0, -padx => 10, -text => $statString)
            ->form(-top => $packL2, -bottom => 'none',
                        -left => $statL1, -right => '%100');
    }

    # show dialog
    my $button = $dlg->Show;

    if ($button eq "Ok") {
        return 0;
    }
    return 1;
}


# Display files from documentation directory in a separate window
sub showDoc {
    my $file = $_[0];
    ::log("Displaying documentation file '$file'\n", 0);
    my $doc = $::DocRoot . $::sep . $file;
    open(DOC, "< $doc") or do {
        ::log("ERROR opening documentation file '$doc': $!\n", 0);
        return 1;
    };
    binmode(DOC);
    my $text = "";
    my $line;
    while ($line = <DOC>) {
        $line = ::chommp($line);
        $text .= $line . "\n";
    }
    close(DOC);

    showDocWindow(" Contents of documentation file '$file' ", $text);
    ::log("...done\n\n", 0);
}


sub showDocWindow {
    my ($title, $text) = @_;
    my $help = $::mainWin->Toplevel(-title => 'aVTC Help');
    # frame for text output
    my $outFrame = $help->LabFrame(@f, -label => $title,
                                       -labelside => "acrosstop");
    my $buttonFrame = $help->Frame(-relief => "flat", -borderwidth => 5);
    # set layout
    $outFrame->form(-top => '%0', -bottom => $buttonFrame,
                    -left => '%0', -right => '%100');
    $buttonFrame->form(-top => 'none', -bottom => '%100',
                       -left => '%0', -right => '%100');
    # exec frame contents
    my $exec = $buttonFrame->Button(-text => 'Close Help Window',
                                    -command => sub { $help->destroy() });
    $exec->pack(-anchor => 'center', -padx => 50, -pady => 2);
    # output frame contents
    my $outScroll = $outFrame->Scrolled('ROText', -scrollbars => "se",
                                    -wrap           => 'none',
                                    -width          => 80,
                                    -height         => 30,
                                    -font           => $fontNorm,
                                    -setgrid        => 1,
    )->pack(-side => 'left', -expand => "yes", -fill => "both");
    my $content = $outScroll->Subwidget('scrolled');
    $content->insert("end", $text);
}


# Display "About" window
sub showAbout {
    my $dlg = $::mainWin->DialogBox(-title => "About", -buttons => ["OK"]);
    $dlg->add("Label", 
	      -text => "aVTC Test Evaluation\n\n$::prog $::version\n".
                       "Graphical user interface $guiVersion",
	      -anchor => "center",
	      -justify => "center",
	      -pady => 10)
        ->pack(@p);
    $dlg->add("Label", 
	      -text => "This program is designed to automate anti-malware ".
	               "and anti-virus software testing.\n".
	               "Functions include test setup and configuration, ".
	               "evaluation, quality control ".
                       "and creation of compressed testbeds.\n",
	      -anchor => "center",
	      -justify => "center",
	      -pady => 10)
        ->pack(@p);
    $dlg->add("Label", 
	      -text => "Copyright (C) 2002-2004 Michel Messerschmidt\n\n".
    "This program is free software; you can redistribute it and/or modify\n".
    "it under the terms of the GNU General Public License as published by\n".
    "the Free Software Foundation; either version 2 of the License, or\n".
    "(at your option) any later version.\n\n".
    "This program is distributed in the hope that it will be useful,\n".
    "but WITHOUT ANY WARRANTY; without even the implied warranty of\n".
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n".
    "GNU General Public License for more details.\n\n".
    "You should have received a copy of the GNU General Public License\n".
    "along with this program; if not, write to the Free Software\n".
    "Foundation, Inc., 59 Temple Place, Suite 330, Boston, ".
    "MA  02111-1307  USA\n",
	      -anchor => "center",
	      -justify => "center",
	      -pady => 10)
	      ->pack(@p);

    $dlg->Show;
}


# Disable or activate widgets as required by currently selected command
sub deactivateWidgets {
    $execWidget->configure(-state => "disabled");
    # disable all selection widgets
    foreach my $opt (keys %activeWidgets) {
        my @widgets = @{$activeWidgets{$opt}};
        foreach my $w (@widgets) {
            $w->configure(-state => "disabled");
        }
    }
    # disable all command widgets
    foreach my $c (keys %cmdWidgets) {
        # currently executed command stays active
        next if ($c eq $::command);
        my $w = $cmdWidgets{$c};
        $w->configure(-state => "disabled");
    }
}


# Disable or activate widgets as required by currently selected command
sub updateActivation {
    my $type = $::cmds{$::command}->{"type"};
    foreach my $opt (keys %activeWidgets) {
        my @widgets = @{$activeWidgets{$opt}};
        if ($type =~ /\Q$opt\E/ and $::loadedTestLists) {
            foreach my $w (@widgets) {
                $w->configure(-state => "normal");
            } 
        } else {
            foreach my $w (@widgets) {
                $w->configure(-state => "disabled");
            }
        }
    }
}


# Disable or activate command widgets (depending on command requirements)
sub updateCommands {
    if ($::debug > 4) {
        ::log("Updating command activation\n", 3, "DEBUG");
    }
    foreach my $c (sort keys %cmdWidgets) {
        my $w = $cmdWidgets{$c};
        $w->configure(-state => "disabled");
        if ($::debug > 4) {
            ::log("Checking command '$c'\n", 3, "DEBUG");
            ::log("Command disabled\n", 4, "DEBUG");
            ::log("cmd requirements: ", 4, "DEBUG");
            foreach my $i (keys %{$::cmds{$c}}) {
                if ($i =~/need|admin/i) {
                    ::log($i.", ");
                }
            }
            ::log("\n");
        }
        if ($::cmds{$c}->{"admin"} and not $adminMode) {
            if ($::debug > 4) {
                ::log("command '$c' requires admin mode -> not activated\n", 
                       5, "DEBUG");
            }
            next;
        }
        if ($::cmds{$c}->{"needsPackCreate"}) {
            if ($::debug > 4) {
                ::log("check requirement: needsPackCreate\n", 4, "DEBUG");
            }
            if ($::loadedPackCreate) {
                $w->configure(-state => "normal");
                if ($::debug > 4) {
                    ::log("activated: loadedpackCreate\n", 5, "DEBUG");
                }
            }
        }
        elsif ($::cmds{$c}->{"needsSum"}) {
            if ($::debug > 4) {
                ::log("check requirement: needsSum\n", 4, "DEBUG");
            }
            if ($::loadedSum) {
                $w->configure(-state => "normal");
                if ($::debug > 4) {
                    ::log("activated: loadedSum\n", 5, "DEBUG");
                }
            }
        }
        elsif ($::cmds{$c}->{"needsNoConfig"}) {
            if ($::debug > 4) {
                ::log("check requirement: needsNoConfig\n", 4, "DEBUG");
            }
            $w->configure(-state => "normal");
            if ($::debug > 4) {
                ::log("activated: needsNoConfig\n", 5, "DEBUG");
            }
        }
        elsif ($::loadedTestConfig and $::loadedTestLists) {
            # there is no option for this command type as
            # this are the "default options"
            if ($::debug > 4) {
                ::log("check requirement: default\n", 4, "DEBUG");
            }
            $w->configure(-state => "normal");
            if ($::debug > 4) {
                ::log("activated: default requirements\n", 5, "DEBUG");
            }
        }
        elsif ($::cmds{$c}->{"needsNoGlobal"}) {
            if ($::debug > 4) {
                ::log("check requirement: needsNoGlobal\n", 4, "DEBUG");
            }
            $w->configure(-state => "normal");
            if ($::debug > 4) {
                ::log("activated: needsNoGlobal\n", 5, "DEBUG");
            }
        }
    }

    # make sure that the currently selected command is active
    my $w = $cmdWidgets{$::command};
    my $state = $w->cget('-state');
    if ($state eq "disabled") {
        $::command = "help";
        $cmdTiler->yview("moveto", 0);
    }
}


# Change to another test to work on
sub updateTest {
    # read new test configuration
    updateLists();

    # update 'selected test' display
    my $title = $::configTestName;
    if ($::loadedTestConfig) {
        $title = $::TestName . "   [" . $::configTestName . "]";
    } else {
        $title = '-- no test selected --';
    }
    $titleLabel->configure(-text => $title);
    # delete old popup help ...
    $popHelp->destroy();
    # ...and create new ones
    $popHelp = $::mainWin->Balloon(-state => 'balloon',
                                -balloonposition => 'mouse',
                                -font => $fontPop);
    # delete old selection widgets
    my @selections = $selFrame->children();
    foreach my $frame (@selections) {
        $frame->destroy();
    }

    # reset widget activation lists
    %activeWidgets = ();
    foreach my $t ("o", "t", "p", "a", "s") {
        @{$activeWidgets{$t}} = ();
    }

    # create new selection widgets
    # This creates the statuscode frame
    my $statusFrame = $selFrame->LabFrame(@f, -label => "Statuscode",
                                        -labelside => "acrosstop");
    $statusFrame->form(-top => '%0', -left => '%0', -right => '%100',
                       -bottom => 'none');
    # remember previous reset status (to workaround strange Optionmenu
    # behaviour: it always applies the last added value)
    my $origResetStatus = $::resetStatus;
    my $statusList = $statusFrame->Optionmenu(-variable => \$::resetStatus);
    foreach my $s (reverse sort keys %::statusCodes) {
        # skip default value and add it as last value
        if ($s eq $origResetStatus) { next; }
        ### Would be nice to have tooltip help here
        ### But they're not supported by the Optionmenu widget
#        my $msg = sprintf("%10s (%s)", $s, $::statusCodes{$s});
        my $msg = $s;
        $statusList->addOptions([$msg => $s]);
    }
#    my $msg = sprintf("%10s (%s)", $origResetStatus,
#                                      $::statusCodes{$origResetStatus});
    my $msg = $origResetStatus;
    $statusList->addOptions([$msg => $origResetStatus]);
    push @{$activeWidgets{"s"}}, $statusList;
    $statusList->form(-top => '%0', -left => '%0', -right => '%100',
                      -bottom => 'none');

    # This creates the operating systems frame
    my $osFrame = $selFrame->LabFrame(@f, -label => "Systems",
                                      -labelside => "acrosstop");
    $osFrame->form(-top => $statusFrame, -left => '%0', -bottom => '%100');
    my %osBox = ();
    $osBox{"button_ALL"} = $osFrame->Button(
                                 -text => "All",
                                 -command => sub { 
                                        foreach my $opt (keys %selectOs) {
                                            $selectOs{$opt} = 1;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"o"}}, $osBox{"button_ALL"};
    $popHelp->attach($osBox{"button_ALL"}, -msg => "select all systems");
    $osBox{"button_NONE"} = $osFrame->Button(
                                 -text => "None",
                                 -command => sub { 
                                        foreach my $opt (keys %selectOs) {
                                            $selectOs{$opt} = 0;
                                        }
                                  },
                                 -relief => 'groove', 
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"o"}}, $osBox{"button_NONE"};
    $popHelp->attach($osBox{"button_NONE"}, -msg => "de-select all systems");
    my $osBoxFrame = $osFrame->Scrolled('Pane',  Name => 'Systeme',
                                        -scrollbars => 'oe',
                                        -sticky => 'nswe');
    $osBoxFrame->form(-top => '%0', -left => '%0', -right => '%100', 
                      -bottom => $osBox{"button_ALL"});
    $osBox{"button_ALL"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => $osBox{"button_NONE"});
    $osBox{"button_NONE"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => '%100');
    my $prevBut = "";
    foreach my $i (sort keys %::osListAll) {
        $osBox{"$i"} = $osBoxFrame->Checkbutton(@cf,
                                    -text => "$i",
                                    -variable => \$selectOs{$i},
                                    -anchor => 'nw',
                                    -justify => 'left');
        if ($prevBut eq "") {
            $osBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => '%0');
        } else {
            $osBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => $prevBut);
        }
        $prevBut = $osBox{"$i"};
        push @{$activeWidgets{"o"}}, $osBox{"$i"};
        $popHelp->attach($osBox{"$i"}, -msg => \$::osListAll{"$i"});
    }

    # This creates the testbeds frame
    my $tbFrame = $selFrame->LabFrame(@f, -label => "Testbeds",
                                      -labelside => "acrosstop");
    $tbFrame->form(-top => $statusFrame, -left => $osFrame, -bottom => '%100');
    my %tbBox = ();
    $tbBox{"button_ALL"} = $tbFrame->Button(
                                 -text => "All",
                                 -command => sub { 
                                        foreach my $opt (keys %selectTb) {
                                            $selectTb{$opt} = 1;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"t"}}, $tbBox{"button_ALL"};
    $popHelp->attach($tbBox{"button_ALL"}, -msg => "select all testbeds");
    $tbBox{"button_NONE"} = $tbFrame->Button(
                                 -text => "None",
                                 -command => sub { 
                                        foreach my $opt (keys %selectTb) {
                                            $selectTb{$opt} = 0;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"t"}}, $tbBox{"button_NONE"};
    $popHelp->attach($tbBox{"button_NONE"}, -msg => "de-select all testbeds");
    my $tbBoxFrame = $tbFrame->Scrolled('Pane',  Name => 'Testbeds',
                                        -scrollbars => 'oe',
                                        -sticky => 'nswe');
    $tbBoxFrame->form(-top => '%0', -left => '%0', -right => '%100', 
                      -bottom => $tbBox{"button_ALL"});
    $tbBox{"button_ALL"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => $tbBox{"button_NONE"});
    $tbBox{"button_NONE"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => '%100');
    $prevBut = "";
    foreach my $i (sort keys %::lists) {
        $tbBox{"$i"} = $tbBoxFrame->Checkbutton(@cf,
                                    -text => "$i",
                                    -variable => \$selectTb{$i},
                                    -anchor => 'nw',
                                    -justify => 'left');
        if ($prevBut eq "") {
            $tbBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => '%0');
        } else {
            $tbBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => $prevBut);
        }
        $prevBut = $tbBox{"$i"};
        push @{$activeWidgets{"t"}}, $tbBox{"$i"};
        my $msg = $::listsAll{"$i"}->{"path"}->{$::configSyntax} . $::sep .
                  "  [" . $::listsAll{"$i"}->{"name"} . "]";
        $popHelp->attach($tbBox{"$i"}, -msg => \$msg);
    }

    # This creates the products frame
    my $prodFrame = $selFrame->LabFrame(@f, -label => "Products",
                                      -labelside => "acrosstop");
    $prodFrame->form(-top => $statusFrame, -left => $tbFrame, -bottom => '%100');
    my %prodBox = ();
    $prodBox{"button_ALL"} = $prodFrame->Button(
                                 -text => "All",
                                 -command => sub { 
                                        foreach my $opt (keys %selectProd) {
                                            $selectProd{$opt} = 1;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"p"}}, $prodBox{"button_ALL"};
    $popHelp->attach($prodBox{"button_ALL"}, -msg => "select all products");
    $prodBox{"button_NONE"} = $prodFrame->Button(
                                 -text => "None",
                                 -command => sub { 
                                        foreach my $opt (keys %selectProd) {
                                            $selectProd{$opt} = 0;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"p"}}, $prodBox{"button_NONE"};
    $popHelp->attach($prodBox{"button_NONE"}, -msg => "de-select all products");
    my $prodBoxFrame = $prodFrame->Scrolled('Pane',  Name => 'Produkte',
                                        -scrollbars => 'oe',
                                        -sticky => 'nswe');
    $prodBoxFrame->form(-top => '%0', -left => '%0', -right => '%100', 
                      -bottom => $prodBox{"button_ALL"});
    $prodBox{"button_ALL"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => $prodBox{"button_NONE"});
    $prodBox{"button_NONE"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => '%100');
    $prevBut = "";
    foreach my $i (sort keys %::prodListAll) {
        $prodBox{"$i"} = $prodBoxFrame->Checkbutton(@cf,
                                    -text => "$i",
                                    -variable => \$selectProd{$i},
                                    -anchor => 'nw',
                                    -justify => 'left');
        if ($prevBut eq "") {
            $prodBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => '%0');
        } else {
            $prodBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => $prevBut);
        }
        $prevBut = $prodBox{"$i"};
        push @{$activeWidgets{"p"}}, $prodBox{"$i"};
        $popHelp->attach($prodBox{"$i"}, -msg => \$::prodListAll{"$i"});
    }

    # selection of pack programs is only used for 'create' command
    my $packFrame = $selFrame->LabFrame(@f, -label => "Packer",
                                      -labelside => "acrosstop");
    $packFrame->form(-top => $statusFrame, -left => $prodFrame,
                     -right => 'none', -bottom => '%100');
    my %packBox = ();
    $packBox{"button_ALL"} = $packFrame->Button(
                                 -text => "All",
                                 -command => sub { 
                                        foreach my $opt (keys %selectPack) {
                                            $selectPack{$opt} = 1;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"a"}}, $packBox{"button_ALL"};
    $popHelp->attach($packBox{"button_ALL"}, -msg => "select all compression programs");
    $packBox{"button_NONE"} = $packFrame->Button(
                                 -text => "None",
                                 -command => sub { 
                                        foreach my $opt (keys %selectPack) {
                                            $selectPack{$opt} = 0;
                                        }
                                  },
                                 -relief => 'groove',
                                 -padx => 0, -pady => '0',
                                 -anchor => 'center',
                                 -justify => 'center');
    push @{$activeWidgets{"a"}}, $packBox{"button_NONE"};
    $popHelp->attach($packBox{"button_NONE"}, -msg => "de-select all compression programs");
    my $packBoxFrame = $packFrame->Scrolled('Pane',  Name => 'Packer',
                                        -scrollbars => 'oe',
                                        -sticky => 'nswe');
    $packBoxFrame->form(-top => '%0', -left => '%0', -right => '%100', 
                      -bottom => $packBox{"button_ALL"});
    $packBox{"button_ALL"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => $packBox{"button_NONE"});
    $packBox{"button_NONE"}->form(-top => 'none', 
                                -left => '%0', -right => '%100', 
                                -bottom => '%100');
    $prevBut = "";
    foreach my $i (sort keys %selectPack) {
        if ($i eq "ALL") { next; }
        $packBox{"$i"} = $packBoxFrame->Checkbutton(@cf,
                                    -text => "$i",
                                    -variable => \$selectPack{$i},
                                    -anchor => 'nw',
                                    -justify => 'left');
        if ($prevBut eq "") {
            $packBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => '%0');
        } else {
            $packBox{"$i"}->form(-left => '%0', -right => '%100',
                               -bottom => 'none', -top => $prevBut);
        }
        $prevBut = $packBox{"$i"};
        push @{$activeWidgets{"a"}}, $packBox{"$i"};
        $popHelp->attach($packBox{"$i"}, -msg => \$::packListAll{"$i"}->{"name"});
    }


    # set active/disabled state for commands
    updateCommands();

    # set active/disabled state for selection widgets
    updateActivation();
}


# read configuration of the currently selected test
sub updateLists {
    my $ret = 0;
    # delete all list elements
    %selectOs = ();
    %selectTb = ();
    %selectProd = ();
    %selectPack = ();

    ::log("Selected test is '$::configTestName'\n", 0);
    ::log("Reading new test configuration...\n", 0);

    # get new test-specific configuration file
    $::configTestFile = $::configAllTests{$::configTestName};

    # read configuration for default test
    ::readTestConfig() or do {
        ::log("\nERROR reading test configuration\n\n", 0);
        $ret = 1;
    };
    if ($::loadedTestConfig) {
        # get lists of all systems/testbeds/products/...
        ::getTestLists() or do {
            ::log("\nERROR reading test configuration lists\n\n", 0);
            $ret = 1;
        };

        if ($::loadedTestLists) {
            foreach (keys %::osListAll) {
                $selectOs{$_} = 0;
            }
            foreach (keys %::listsAll) {
                $selectTb{$_} = 0;
            }
            foreach (keys %::prodListAll) {
                $selectProd{$_} = 0;
            }
            foreach (keys %::packListAll) {
                $selectPack{$_} = 0;
            }
            ::log("...done\n", 0);
        }
    } else {
        # reset flags
        $::loadedTestLists = 0;
        $::loadedSum = 0;
        $::loadedPack = 0;
        $::loadedPackCreate = 0;
        $::needsPack = 0;
        # delete old lists
        %::osListAll = ();
        %::osList = ();
        %::listsAll = ();
        %::lists = ();
        %::prodListAll = ();
        %::prodList = ();
        @::noTestListAll = ();
        %::packListAll = ();
        %::packList = ();
    }

    return $ret;
}


# return a true value
1;
