# Ranking.pm
# implements 'report rank' command
#
# v0.17-vir-both, 2003-05-26
# Copyright (c) Michel Messerschmidt 2002,2003
#
package Avtc::Ranking;
use strict;
use Exporter;
use Text::Wrap;

@Avtc::Ranking::ISA = qw(Exporter);
@Avtc::Ranking::EXPORT = qw(reportRank);


sub reportRank {
    if ($::TestType =~ /^heureka$/io) {
        return doRankingHeureka();
    } else {
        ::log("This command is available for Heureka tests only.\n", 1);
        return 1;
    }
}


sub doRankingHeureka {
    my $ret = 0;
    if ($::debug) {
        print($::DebugFile "ranking heureka test:\n");
        print("ranking heureka test:\n");
    }
    my (%prefixList, $listNr, %data, %dataSum, %dataRank, %rankCrit);
    # documentation on data structures:
    #
    # %data holds most of the data: it's a hash with many dimensions
    # first key is the product: %data{"FPR}, %data{"AVP"},...
    # second key is the operating system (only one at the moment):
    #   %data{$prod}->{"WNT"}
    # third key is either the evaluation list:
    #       $list      (eg. %data{$prod}->{$os}->{"SZ___"})
    # or    "rank"     -  the overall rank of this product as integer
    # or    "points"   -  the total number of points according to the ranks
    #                     achieved in each list/testbed
    #
    # the fourth key is more complicated: %data{$prod}->{$os}->{$list}->{$key4}
    # with $key4 =  "status"  - this contains either a " " or a "*":
    #                           " " means this products test is finished
    #                           "*" means that testing/evaluation is in work,
    #                               therefore the result may not be final !!!
    #               "ref"     -  the data from the reference test is here
    #               "3mon"    -  the data from the 3 month later testbeds
    #               "6mon"    -  the data from the 6 month later testbeds
    #                         each of this 3 keys values is a reference to
    #                         an array containing the real data:
    #                            [0] is not used (to avoid index translations)
    #                            [1] is the total number of viruses found
    #                            ($data{$prod}->{$os}->{$list}->{"ref"}->[1])
    #                            [2] is the percent value of this total number
    #                            [3] is the total nr of unreliable identified viruses
    #                            [4] is the percent value of this
    #                            [5] is the total nr of unreliable detected viruses
    #                            [6] is the percent value of this
    #                            [7] is the total number of files found
    #                            [8] is the percent value of this
    #               "loss3"   -  the percentage loss after the first 3 months
    #               "loss6"   -  the percentage loss after 6 months
    #               "rank"    -  the rank of this product in this testbed/list
    #                            as integer
    #
    #
    # %dataSum contains the total number of viruses in each testbed
    # in the similar hierachy to the corresponding values in %data:
    #  [0] is the total number and percentage of viruses found:
    #        $dataSum{$os}->{$list}->{"ref"}->[0]
    #        $dataSum{$os}->{$list}->{"3mon"}->[0]
    #        $dataSum{$os}->{$list}->{"6mon"}->[0]
    #  [1] is the total number of viruses found:
    #  [2] is the percent value of viruses found:
    #  [3] is the total number and percentage of files found
    #  [4] is the total number of files found
    #  [5] is the percent value of files found
    # (Attention: there IS a difference: the values stored in [0] and [3] are
    #             strings containing the number of viruses and the percent
    #             value, which is of cource always 100.0%)
    #
    #
    # %dataRank contains lists of products with the same rank
    #  $dataRank{"x"}->{"name"}        - with x=1-5 contains name of this rank
    #  $dataRank{"x"}->{$os}->{"prods"} - with x=1-5 contains all products with
    #                                    an overall rank of x
    #  $dataRank{"x"}->{$os}->{$list}  - with x=1-5 contains all products with
    #                                    this rank in this testbed/list
    #
    # %rankCrit contains rabnking criterias for the different testbeds
    # $rankCrit{tbType}->{rank} = value   - where value is the minimum percent
    #                                       value for this rank
    #
    $dataRank{"5"}->{"name"} = "Perfect";
    $dataRank{"4"}->{"name"} = "Excellent";
    $dataRank{"3"}->{"name"} = "Very good";
    $dataRank{"2"}->{"name"} = "Good";
    $dataRank{"1"}->{"name"} = "Acceptable";
    $dataRank{"0"}->{"name"} = "Useless";

    $rankCrit{"I"}->{"5"} = 100.0;
    $rankCrit{"I"}->{"4"} =  99.0;
    $rankCrit{"I"}->{"3"} =  95.0;
    $rankCrit{"I"}->{"2"} =  90.0;
    # can never be reached - no points for this in ITW ranking
    $rankCrit{"I"}->{"1"} =  90.0;
    $rankCrit{"I"}->{"0"} =   0;

    $rankCrit{"Z"}->{"5"} = 100.0;
    $rankCrit{"Z"}->{"4"} =  99.0;
    $rankCrit{"Z"}->{"3"} =  95.0;
    $rankCrit{"Z"}->{"2"} =  90.0;
    $rankCrit{"Z"}->{"1"} =  80.0;
    $rankCrit{"Z"}->{"0"} =   0;

    $rankCrit{"M"}->{"5"} = 100.0;
    $rankCrit{"M"}->{"4"} =  90.0;
    $rankCrit{"M"}->{"3"} =  80.0;
    $rankCrit{"M"}->{"2"} =  70.0;
    $rankCrit{"M"}->{"1"} =  60.0;
    $rankCrit{"M"}->{"0"} =   0;

    my $os;
    LOOP_OS: foreach $os (keys %::osList) {
        # get a list of all variantlists without the month indicator characters
        my ($temp, $tempList, $tempType, $prod);
        $listNr = 0;
        foreach $tempList (keys %::listsAll) {
            # tempType is Z, I, M, ... = Zoo, ITW, Malware, ...
            $tempType = substr($tempList, 1, 1);
            # remove the last 3 characters (denoting the month)
            substr($tempList, -3) = "";
            if (not defined ($prefixList{$tempList})) {
                $prefixList{$tempList} = $tempType;
                $listNr++;
            }
        }

        my $list;
        LOOP_LIST: foreach $list (keys %prefixList) {
            print($::DebugFile "processing list: ",$list,"  -  os: ",$os,"\n");
            print("processing list: ",$list,"  -  os: ",$os,"\n");
            # define private variables used in this block
            my $line;
            my $dataRefFile = $::EvalBase;
            # cut TestName from Dir and add RefTestName
            $dataRefFile =~ s-\/[^/]+$-\/-o;
            $dataRefFile .= $::RefTestName . "/result/" . $list . "___." . $os;
            my $data3MonFile = $::ResultDir . "/" . $list . $::TestMonth3Name .
                               "." . $os;
            my $data6MonFile = $::ResultDir . "/" . $list . $::TestMonth6Name .
                               "." . $os;
            if ($::debug) {
                print($::DebugFile "DataRefFile = ",$dataRefFile,"\n");
                print($::DebugFile "Data3MonFile = ",$data3MonFile,"\n");
                print($::DebugFile "Data6MonFile = ",$data6MonFile,"\n");
            }

            # get data from reference test
            open(DATAREF, "< $dataRefFile") or do {
                print($::DebugFile "ERROR reading result file: $!\n");
                print($::DebugFile "      !!! Ranking generation of ",$dataRefFile," stopped !!!\n\n");
                print("ERROR reading result file: $!\n");
                print("      !!! Ranking generation of ",$dataRefFile," stopped !!!\n\n");
                close(DATAREF);
                $ret = 1;
                next LOOP_LIST;
            };
            binmode(DATAREF);
            while (defined($line = <DATAREF>)) {
                $line = ::chommp($line);
                if ($line =~ /^Testbed/io) {
                    # get the Testbed line with total numbers
                    my @tempField = split(/\s+/, $line);
                    $dataSum{$os}->{$list}->{"ref"}->[0] =
                            $tempField[1] . " " . $tempField[2];
                    $dataSum{$os}->{$list}->{"ref"}->[1] = $tempField[1];
                    $dataSum{$os}->{$list}->{"ref"}->[2] = $tempField[2];
                    $dataSum{$os}->{$list}->{"ref"}->[3] =
                            $tempField[3] . " " . $tempField[4];
                    $dataSum{$os}->{$list}->{"ref"}->[4] = $tempField[3];
                    $dataSum{$os}->{$list}->{"ref"}->[5] = $tempField[4];
                } elsif ($line =~ /^[A-Z0-9][A-Z0-9][A-Z0-9] /io ) {
                    # get all scanner data lines
                    ### WARNING: this doesn't work with tabs
                    my @tempField = split(/\s+/, $line);
                    # test if is it valid data or a status message
                    if ($tempField[1] =~ /^\*/o) {
                        $tempField[1] = "---";
                        $tempField[2] = 0;
                        $tempField[3] = "---";
                        $tempField[4] = 0;
                        $tempField[5] = "---";
                        $tempField[6] = 0;
                        $tempField[7] = "---";
                        $tempField[8] = 0;
                    }
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[0] = "";
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[1] = $tempField[1];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[2] = $tempField[2];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[3] = $tempField[3];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[4] = $tempField[4];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[5] = $tempField[5];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[6] = $tempField[6];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[7] = $tempField[7];
                    $data{$tempField[0]}->{$os}->{$list}->{"ref"}->[8] = $tempField[8];
                    # status can be ' ' for ok or '*' for in work
                    # reference data is of cource always ok
                    $data{$tempField[0]}->{$os}->{$list}->{"status"} = " ";
                }
                # ignore all other lines
            }
            close(DATAREF);

            # get data from 3month testbeds
            open(DATA3MON, "< $data3MonFile") or do {
                print($::DebugFile "ERROR reading result file: $!\n");
                print($::DebugFile "      !!! Ranking generation of ",$data3MonFile," stopped !!!\n\n");
                print("ERROR reading result file: $!\n");
                print("      !!! Ranking generation of ",$data3MonFile," stopped !!!\n\n");
                close(DATA3MON);
                $ret = 1;
                next LOOP_LIST;
            };
            binmode(DATA3MON);

            while (defined($line = <DATA3MON>)) {
                $line = ::chommp($line);
                if ($line =~ /^Testbed/io) {
                    my @tempField = split(/\s+/, $line);
                    $dataSum{$os}->{$list}->{"3mon"}->[0] =
                            $tempField[1] . " " . $tempField[2];
                    $dataSum{$os}->{$list}->{"3mon"}->[1] = $tempField[1];
                    $dataSum{$os}->{$list}->{"3mon"}->[2] = $tempField[2];
                    $dataSum{$os}->{$list}->{"3mon"}->[3] =
                            $tempField[3] . " " . $tempField[4];
                    $dataSum{$os}->{$list}->{"3mon"}->[4] = $tempField[3];
                    $dataSum{$os}->{$list}->{"3mon"}->[5] = $tempField[4];
                } elsif ($line =~ /^[A-Z0-9][A-Z0-9][A-Z0-9]\*? /io ) {
                    my @tempField = split(/\s+/, $line);
                    # remove '*' from product name
                    if (length($tempField[0]) == 4) {
                        chop($tempField[0]);
                        $data{$tempField[0]}->{$os}->{$list}->{"status"} = "*";
                    }
                    # test if is it valid data or a status message
                    if ($tempField[1] =~ /^\*/o) {
                        $tempField[1] = "---";
                        $tempField[2] = 0;
                        $tempField[3] = "---";
                        $tempField[4] = 0;
                        $tempField[5] = "---";
                        $tempField[6] = 0;
                        $tempField[7] = "---";
                        $tempField[8] = 0;
                    }
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[0] = "";
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[1] = $tempField[1];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[2] = $tempField[2];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[3] = $tempField[3];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[4] = $tempField[4];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[5] = $tempField[5];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[6] = $tempField[6];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[7] = $tempField[7];
                    $data{$tempField[0]}->{$os}->{$list}->{"3mon"}->[8] = $tempField[8];
                }
                # ignore all other lines
            }
            close(DATA3MON);

            # get data from 6 month testbeds
            open(DATA6MON, "< $data6MonFile") or do {
                print($::DebugFile "ERROR reading result file: $!\n");
                print($::DebugFile "      !!! Ranking generation of ",$data6MonFile," stopped !!!\n\n");
                print("ERROR reading result file: $!\n");
                print("      !!! Ranking generation of ",$data6MonFile," stopped !!!\n\n");
                close(DATA3MON);
                $ret = 1;
                next LOOP_LIST;
            };
            binmode(DATA6MON);
            while (defined($line = <DATA6MON>)) {
                $line = ::chommp($line);
                if ($line =~ /^Testbed/io) {
                    my @tempField = split(/\s+/, $line);
                    $dataSum{$os}->{$list}->{"6mon"}->[0] =
                            $tempField[1] . " " . $tempField[2];
                    $dataSum{$os}->{$list}->{"6mon"}->[1] = $tempField[1];
                    $dataSum{$os}->{$list}->{"6mon"}->[2] = $tempField[2];
                    $dataSum{$os}->{$list}->{"6mon"}->[3] =
                            $tempField[3] . " " . $tempField[4];
                    $dataSum{$os}->{$list}->{"6mon"}->[4] = $tempField[4];
                    $dataSum{$os}->{$list}->{"6mon"}->[5] = $tempField[5];
                } elsif ($line =~ /^[A-Z0-9][A-Z0-9][A-Z0-9]\*? /io ) {
                    my @tempField = split(/\s+/, $line);
                    # remove '*' from product name
                    if (length($tempField[0]) == 4) {
                        chop($tempField[0]);
                        $data{$tempField[0]}->{$os}->{$list}->{"status"} = "*";
                    }
                    # test if is it valid data or a status message
                    if ($tempField[1] =~ /^\*/o) {
                        $tempField[1] = "---";
                        $tempField[2] = 0;
                        $tempField[3] = "---";
                        $tempField[4] = 0;
                        $tempField[5] = "---";
                        $tempField[6] = 0;
                        $tempField[7] = "---";
                        $tempField[8] = 0;
                    }
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[0] = "";
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[1] = $tempField[1];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[2] = $tempField[2];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[3] = $tempField[3];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[4] = $tempField[4];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[5] = $tempField[5];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[6] = $tempField[6];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[7] = $tempField[7];
                    $data{$tempField[0]}->{$os}->{$list}->{"6mon"}->[8] = $tempField[8];
                }
                # ignore all other lines
            }
            close(DATA6MON);

            # remove all product data if it was not tested in all tests
            my %delList = ();
            foreach (keys %data) {
                if (not exists ($data{$_}->{$os}->{$list}->{"3mon"}) or
                    not exists ($data{$_}->{$os}->{$list}->{"6mon"}) or
                    not exists ($data{$_}->{$os}->{$list}->{"ref"})) {
                    $delList{$_} = $_;
                    if ($::debug) {
                        print($::DebugFile "deleting ",$_," from product list\n");
            }
                }
            }
            foreach (keys %delList) {
               delete $data{$_};
            }

            if ($::debug) {
                print($::DebugFile "DataRefProds = ");
                foreach (sort keys %data) { print(" ",$_); }
                print("\n");
            }

            foreach $prod (keys %data) {
                my $tempVal = -1 * (
                                $data{$prod}->{$os}->{$list}->{"ref"}->[2]
                              - $data{$prod}->{$os}->{$list}->{"3mon"}->[2]);
#                if ($tempVal > 0) {
#                    $data{$prod}->{$os}->{$list}->{"loss3"} = "+";
#                } else {
#                    $data{$prod}->{$os}->{$list}->{"loss3"} = "";
#                }
                $data{$prod}->{$os}->{$list}->{"loss3"} =
                        sprintf("%6.1f", $tempVal);

                $tempVal = -1 * (
                             $data{$prod}->{$os}->{$list}->{"ref"}->[2]
                           - $data{$prod}->{$os}->{$list}->{"6mon"}->[2]);
#                if ($tempVal > 0) {
#                    $data{$prod}->{$os}->{$list}->{"loss3"} = "+";
#                } else {
#                    $data{$prod}->{$os}->{$list}->{"loss3"} = "";
#                }
                $data{$prod}->{$os}->{$list}->{"loss6"} .=
                        sprintf("%6.1f", $tempVal);
            }


            #write compare lists to file (for each list)
            my $resultFile = $::ReportDir . "/" . $list . "CMP." . $os;
            open(RESULT, "> $resultFile") or do {
                print($::DebugFile "ERROR creating $resultFile: $!\n");
                print($::DebugFile "      !!! Ranking of ",$list," stopped !!!\n\n");
                print("ERROR creating $resultFile: $!\n");
                print("      !!! Ranking of ",$list," stopped !!!\n\n");
                close(RESULT);
                $ret = 1;
                next LOOP_LIST;
            };
            binmode(RESULT);

            print(RESULT "Test           ",$::RefTestName,"  |       ",
                         $::TestName,"-",$::TestMonth3Name,"       |       ",
                         $::TestName,"-",$::TestMonth6Name," ", $::EOL);
#            print(RESULT "------------------------+----------------+----------------", $::EOL);
#            print(RESULT "             Viruses    |   New viruses  |   New viruses", $::EOL);
#            print(RESULT "Scanner      detected   |    detected    |    detected", $::EOL);
#            print(RESULT "------------------------+----------------+----------------", $::EOL);
            print(RESULT "------------------------+---------------+---------+---------------+----------", $::EOL);
            print(RESULT "             Viruses    |   New viruses | loss in |   New viruses | loss in ", $::EOL);
            print(RESULT "Scanner      detected   |    detected   | 3 months|    detected   | 6 months", $::EOL);
            print(RESULT "------------------------+---------------+---------+---------------+----------", $::EOL);
            printf(RESULT "Testbed   %12s  |  %12s           |  %12s%s",
                          $dataSum{$os}->{$list}->{"ref"}->[0],
                          $dataSum{$os}->{$list}->{"3mon"}->[0],
                          $dataSum{$os}->{$list}->{"6mon"}->[0], $::EOL);
            print(RESULT "------------------------+---------------+---------+---------------+----------", $::EOL);
            foreach $prod (sort keys %data) {
                printf(RESULT "%s%s      %5s %5s%%  |  %5s %5s%% | %6s%% |  %5s %5s%% | %6s%%%s", $prod,
                          $data{$prod}->{$os}->{$list}->{"status"},
                          $data{$prod}->{$os}->{$list}->{"ref"}->[1],
                          $data{$prod}->{$os}->{$list}->{"ref"}->[2],
                          $data{$prod}->{$os}->{$list}->{"3mon"}->[1],
                          $data{$prod}->{$os}->{$list}->{"3mon"}->[2],
                          $data{$prod}->{$os}->{$list}->{"loss3"},
                          $data{$prod}->{$os}->{$list}->{"6mon"}->[1],
                          $data{$prod}->{$os}->{$list}->{"6mon"}->[2],
                          $data{$prod}->{$os}->{$list}->{"loss6"}, $::EOL);
            }
            close(RESULT);

            #write percentage loss lists to file (for each list)
#            my $resultFile2 = $::ReportDir . "/" . $list . "LOS." . $os;
#            open(RESULT2, "> $resultFile2") or do {
#                print("ERROR creating $resultFile2: $!\n");
#                print("      !!! Ranking of ",$list," stopped !!!\n\n");
#                close(RESULT2);
#                next LOOP_LIST;
#            };
#            binmode(RESULT2);

#            print(RESULT2 "----------------------------------------------------------", $::EOL);
#            print(RESULT2 "         detection rate | loss after | loss after", $::EOL);
#            print(RESULT2 "Scanner   in ref-test   |  3 months  |  6 months", $::EOL);
#            print(RESULT2 "------------------------+------------+--------------------", $::EOL);
#            foreach $prod (sort keys %data) {
#                printf(RESULT2 "%s%s          %5s%%    |  %5s%%   |  %5s%%%s", $prod,
#                          $data{$prod}->{$os}->{$list}->{"status"},
#                          $data{$prod}->{$os}->{$list}->{"ref"}->[2],
#                          $data{$prod}->{$os}->{$list}->{"loss3"},
#                          $data{$prod}->{$os}->{$list}->{"loss6"}, $::EOL);
#            }
#            close(RESULT2);

        } # end of LOOP_LIST


        # Do the product ranking based on the percentage of found viruses.
        ### Not true anymore
        # Do the product ranking based on the percentage loss.
        ###
        # Ranks in the testreport ("perfect", "excellent", ...) are
        # represented by integer values here, where 1 is the lowest rank,
        # 2 the second lowest rank, ... and 5 is the highest rank
        # (this may change)
        if ($::debug) {
            print($::DebugFile "begin ranking...\n");
        }
        foreach $temp (keys %dataRank) {
            $dataRank{$temp}->{$os}->{"prods"} = "";
            foreach $list (keys %prefixList) {
                $dataRank{$temp}->{$os}->{$list} = "";
            }
        }
        foreach $prod (sort keys %data) {
            # 99 is no real rank, just an inital value
            my $tempRank = 99;
            $data{$prod}->{$os}->{"points"} = 0;
            # now find the lowest rank from all evaluation lists
            foreach $list (keys %prefixList) {
                my $type = $prefixList{$list};
                # make the next lines more readable
                my $refV = $data{$prod}->{$os}->{$list}->{"ref"}->[2];
                my $mon3V = $data{$prod}->{$os}->{$list}->{"3mon"}->[2];
                my $mon6V = $data{$prod}->{$os}->{$list}->{"6mon"}->[2];
#                my $refF = $data{$prod}->{$os}->{$list}->{"ref"}->[8];
#                my $loss3 = $data{$prod}->{$os}->{$list}->{"loss3"};
#                my $loss6 = $data{$prod}->{$os}->{$list}->{"loss6"};
                if ($::debug) {
                    print($::DebugFile "prod=",$prod,"   -  list=",
                          $list,"   -   os=",$os,"\n");
                }
                if ($mon3V >= $rankCrit{$type}->{"5"}
                    and $mon6V >= $rankCrit{$type}->{"5"}) {
                    $tempRank = 5 unless ($tempRank < 5);
                    $data{$prod}->{$os}->{$list}->{"rank"} = 5;
                    $data{$prod}->{$os}->{"points"} += 2;
                    $dataRank{"5"}->{$os}->{$list} .= $prod . ", ";
                    if ($::debug) {
                        print($::DebugFile "SETRANK: tempRank=",$tempRank,
                              "  -  rank=5  -  points=",
                              $data{$prod}->{$os}->{"points"},"\n");
                    }
                } elsif ($mon3V >= $rankCrit{$type}->{"4"} and
                         $mon6V >= $rankCrit{$type}->{"4"}) {
                    $tempRank = 4 unless ($tempRank < 4);
                    $data{$prod}->{$os}->{$list}->{"rank"} = 4;
                    $data{$prod}->{$os}->{"points"} += 1;
                    $dataRank{"4"}->{$os}->{$list} .= $prod . ", ";
                    if ($::debug) {
                        print($::DebugFile "SETRANK: tempRank=",$tempRank,
                              "  -  rank=4  -  points=",
                              $data{$prod}->{$os}->{"points"},"\n");
                    }
                } elsif ($mon3V >= $rankCrit{$type}->{"3"} and
                         $mon6V >= $rankCrit{$type}->{"3"}) {
                    $tempRank = 3 unless ($tempRank < 3);
                    $data{$prod}->{$os}->{$list}->{"rank"} = 3;
#                    $data{$prod}->{$os}->{"points"} += 3;
                    $dataRank{"3"}->{$os}->{$list} .= $prod . ", ";
                    if ($::debug) {
                        print($::DebugFile "SETRANK: tempRank=",$tempRank,
                              "  -  rank=3  -  points=",
                              $data{$prod}->{$os}->{"points"},"\n");
                    }
                } elsif ($mon3V >= $rankCrit{$type}->{"2"} and
                         $mon6V >= $rankCrit{$type}->{"2"}) {
                    $tempRank = 2 unless ($tempRank < 2);
                    $data{$prod}->{$os}->{$list}->{"rank"} = 2;
#                    $data{$prod}->{$os}->{"points"} += 2;
                    $dataRank{"2"}->{$os}->{$list} .= $prod . ", ";
                    if ($::debug) {
                        print($::DebugFile "SETRANK: tempRank=",$tempRank,
                              "  -  rank=2  -  points=",
                              $data{$prod}->{$os}->{"points"},"\n");
                    }
                } elsif ($mon3V >= $rankCrit{$type}->{"1"} and
                         $mon6V >= $rankCrit{$type}->{"1"}) {
                    $tempRank = 1 unless ($tempRank < 1);
                    $data{$prod}->{$os}->{$list}->{"rank"} = 1;
#                    $data{$prod}->{$os}->{"points"} += 1;
                    $dataRank{"1"}->{$os}->{$list} .= $prod . ", ";
                    if ($::debug) {
                        print($::DebugFile "SETRANK: tempRank=",$tempRank,
                              "  -  rank=1  -  points=",
                              $data{$prod}->{$os}->{"points"},"\n");
                    }
                } else {
                    $tempRank = 0;
                    $data{$prod}->{$os}->{$list}->{"rank"} = 0;
                    # no points for such results
                    $dataRank{"0"}->{$os}->{$list} .= $prod . ", ";
                    if ($::debug) {
                        print($::DebugFile "SETRANK: tempRank=",$tempRank,
                              "  -  rank=0  -  points=",
                              $data{$prod}->{$os}->{"points"},"\n");
                    }
                }
            }
            $data{$prod}->{$os}->{"rank"} = $tempRank;
            $dataRank{$tempRank}->{$os}->{"prods"} .= $prod . ", ";
        }


        # format output
        $::Text::Wrap::colums = 65;
        foreach $temp (keys %dataRank) {
            if (length($dataRank{$temp}->{$os}->{"prods"}) > 1) {
                # remove last ", "
                chop($dataRank{$temp}->{$os}->{"prods"});
                chop($dataRank{$temp}->{$os}->{"prods"});
            }
            $dataRank{$temp}->{$os}->{"prods"} = $dataRank{$temp}->{"name"}
                    . " Products: " . $dataRank{$temp}->{$os}->{"prods"};
            $dataRank{$temp}->{$os}->{"prods"} = wrap("", "        ",
                    $dataRank{$temp}->{$os}->{"prods"});
            foreach $list (keys %prefixList) {
                if (length($dataRank{$temp}->{$os}->{$list}) > 1) {
                    chop($dataRank{$temp}->{$os}->{$list});
                    chop($dataRank{$temp}->{$os}->{$list});
                }
                $dataRank{$temp}->{$os}->{$list} = $dataRank{$temp}->{"name"}
                    . " Products: " . $dataRank{$temp}->{$os}->{$list};
                $dataRank{$temp}->{$os}->{$list} = wrap("", "        ",
                        $dataRank{$temp}->{$os}->{$list});
            }
        }


        #write ranking lists to file (for each list)
        my $resultFile3 = $::ReportDir . "/RANKING." . $os;
        open(RESULT3, "> $resultFile3") or do {
            print($::DebugFile "ERROR creating $resultFile3: $!\n");
            print($::DebugFile "      !!! Ranking stopped !!!\n\n");
            print("ERROR creating $resultFile3: $!\n");
            print("      !!! Ranking stopped !!!\n\n");
            close(RESULT3);
            $ret = 1;
        };
        binmode(RESULT3);


        print(RESULT3 "--- Overall grading of products for all testbeds -------------------", $::EOL);
        print(RESULT3 $dataRank{"5"}->{$os}->{"prods"}, $::EOL);
        print(RESULT3 $dataRank{"4"}->{$os}->{"prods"}, $::EOL);
        print(RESULT3 $dataRank{"3"}->{$os}->{"prods"}, $::EOL);
        print(RESULT3 "--------------------------------------------------------------------", $::EOL, $::EOL, $::EOL);
        foreach $list (keys %prefixList) {
            print(RESULT3 "--- Grading of products for ",$list," ---------------------------", $::EOL);
            print(RESULT3 $dataRank{"5"}->{$os}->{$list}, $::EOL);
            print(RESULT3 $dataRank{"4"}->{$os}->{$list}, $::EOL);
            print(RESULT3 $dataRank{"3"}->{$os}->{$list}, $::EOL);
            print(RESULT3 "--------------------------------------------------------------------", $::EOL, $::EOL);
        }


        print(RESULT3 $::EOL, $::EOL, $::EOL, $::EOL,
                      "--------------------------------------------------------------------", $::EOL);
        print(RESULT3 "This results in the following ranking (out of ",$listNr * 2," possible points):", $::EOL);
        print(RESULT3 "--------------------------------------------------------------------", $::EOL);

        my %tempHash;
        foreach $prod (keys %data) {
            if ($::debug) {
                print($::DebugFile "prod=",$prod," - points=",
                      $data{$prod}->{$os}->{"points"},"\n");
            }
            my $points = $data{$prod}->{$os}->{"points"};
            if (exists $tempHash{$points}) {
                $tempHash{$points} .= "," . $prod;
            } else {
                $tempHash{$points} = $prod;
            }
        }
        my $key;
        my $counter = 1;
        foreach $key (sort {$b <=> $a} keys(%tempHash)) {
            if ($::debug) {
                print($::DebugFile "sorted points=",$key," - prod=",
                      $tempHash{$key}," - counter=",$counter,"\n");
            }
            printf(RESULT3 " %2d. place (%2s points): %-30s%s", $counter,
                   $key, $tempHash{$key}, $::EOL);
        my @tempProds = split(",", $tempHash{$key});
            $counter += $#tempProds + 1;

        }

        print(RESULT3 "----------------------------------------------------------", $::EOL);
        close(RESULT3);

    } # end of LOOP_OS

    return $ret;
}


# return a true value
1;