# Rpt_info.pm
# implements 'report info' and 'report problems' command
#
# v0.16, 2004-04-27
# Copyright (c) Michel Messerschmidt 2002-2004
#
package Avtc::Rpt_info;
use strict;
use Exporter;
use File::Copy;

@Avtc::Rpt_info::ISA = qw(Exporter);
@Avtc::Rpt_info::EXPORT = qw(reportInfo reportProblems);


sub reportInfo {
    my $ret = 0;
    my $infoHeader = "=======================" . $::EOL .
                     "File " . $::FileReportProdInfo . $::EOL .
                     "=======================" . $::EOL;
    # status filenames are set in test configuration
    my $outfile = $::ReportDir . $::sep . $::FileReportProdInfo;
    open(REPORT, "> $outfile") or do {
        ::log("ERROR: cannot create $outfile: $!\n", 1);
        return 1;
    };
    binmode(REPORT);
    ::log("writing to: $outfile\n", 1);
    print(REPORT $infoHeader, $::EOL, $::EOL, $::EOL);

    my $prod;
    LOOP_PROD: foreach $prod (sort keys %::prodListAll) {
        my $os;
        my $allWin = 0;
        my %options = ();
        my %versions = ();
        my $infoDir = $::InfoRoot.$::sep.$prod;
        if ($::debug > 0) {
            ::log("InfoDir = $infoDir\n", 1, "DEBUG: ");
        }
        ::log("collecting info about $prod...\n", 1);

        printf(REPORT "CodeName:       %s%s", $prod, $::EOL);
        printf(REPORT "Product:        %s%s", $::prodListAll{$prod}, $::EOL);


        print(REPORT "Versions:",$::EOL);
        LOOP_VER: foreach $os (sort keys %::osListAll) {
            # check for not tested combinations
            foreach (@::noTestListAll) {
                if ($_ eq "$os ALL $prod") {
                    if ($::debug > 0) {
                        ::log("$prod not tested on system $os, ".
                              "skipping $os data\n", 2);
                    }
                    next LOOP_VER;
                }
            }
            my $tempLine;
            # handle windows system different (maybe there is just one file
            # for all windows systems (with filename extension "WIN")
            if ($os =~ /^W/i) {
                if ($allWin) {
                    next LOOP_VER;
                } elsif (-e $infoDir . $::sep . $::configDirNames{"infoVer"} .
                            $::sep . $::configTestName . ".WIN") {
                    $allWin = 1;
                    $os = "Win";
                }

            }
            # check for "NO TEST" status on a specific system
            if (checkNoTestStatus($prod, $os) == 1) {
                # skip this OS because all testbeds have "NO TEST" status
                next LOOP_VER;
            }

            my $infoFile = $infoDir . $::sep . $::configDirNames{"infoVer"} .
                           $::sep . $::configTestName . "." . $os;

            if (not -e $infoFile) {
                $versions{$os}->{"prodver"} = "MISSING";
                $versions{$os}->{"proddate"} = "MISSING";
                $versions{$os}->{"defver1"} = "MISSING";
                $versions{$os}->{"defdate1"} = "MISSING";
            } else {
                open(INFO, "< $infoFile") or do {
                    ::log("ERROR: cannot open $infoFile: $!\n", 2);
                    close REPORT;
                    return 1;
                };
                # These keys must exist, so they are initialized here
                # even if they contain no data afterwards
                $versions{$os}->{"prodver"} = "";
                $versions{$os}->{"proddate"} = "";
                $versions{$os}->{"defver1"} = "";
                $versions{$os}->{"defdate1"} = "";
                $versions{$os}->{"defver"} = "";
                $versions{$os}->{"defdate"} = "";
                my $err = 0;
                while (defined($tempLine = <INFO>)) {
                    # ignore comment lines (must begin with a '#', ';' or 'REM')
                    if ($tempLine =~ /^(REM|;|\#)/i) { next; }
                    # ignore empty lines
                    if ($tempLine =~ /^\s*$/) { next; }
                    # ignore lines containing only the product name
                    if ($tempLine =~ /^\s*$prod\s*$/i) { next; }
                    # cut all control chars from end of line
                    $tempLine = ::chommp($tempLine);
                    if ($tempLine =~ /^PRODVER: (.*)/i) {
                        $versions{$os}->{"prodver"} .= $1;
                    }
                    elsif ($tempLine =~ /^PRODDATE: (.*)/i) {
                        $versions{$os}->{"proddate"} .= $1;
                    }
                    elsif ($tempLine =~ /^ENGVER([1-9]): (.*)/i) {
                        my $engine = "engver" . $1;
                        $versions{$os}->{$engine} .= $2;
                    }
                    elsif ($tempLine =~ /^ENGDATE([1-9]): (.*)/i) {
                        my $engine = "engver" . $1;
                        $versions{$os}->{$engine} .= $2;
                    }
                    elsif ($tempLine =~ /^DEFVER([1-9]): (.*)/i) {
                        my $defver = "defver" . $1;
                        $versions{$os}->{$defver} .= $2;
                    }
                    elsif ($tempLine =~ /^DEFDATE([1-9]): (.*)/i) {
                        my $defdate = "defdate" . $1;
                        $versions{$os}->{$defdate} .= $2;
                    }
                    else {
                        if (not $err) {
                            ::log("ERROR parsing version file for $prod ".
                                  "on $os\n", 2);
                            $err = 1;
                            $ret = 1;
                        }
                        if ($::debug > 0) {
                           ::log("unknown line: $tempLine\n", 2, "DEBUG: ");
                        }
                    }
                }
                close INFO;
            }

            printf(REPORT "    %s:", $os);
            printf(REPORT " %s", $versions{$os}->{"prodver"});
            printf(REPORT "  date: %s", $versions{$os}->{"proddate"});
            print(REPORT $::EOL);
            foreach my $i (1..9) {
                my $engver = "engver" . $i;
                my $engdate = "engdate" . $i;
                if (defined($versions{$os}->{$engver}) or
                    defined($versions{$os}->{$engdate})) {
                    printf(REPORT "         Eng:", $os);
                    if (defined $versions{$os}->{$engver}) {
                        printf(REPORT " %s", $versions{$os}->{$engver});
                    }
                    if (defined $versions{$os}->{$engdate}) {
                        printf(REPORT "  date: %s", $versions{$os}->{$engdate});
                    }
                    print(REPORT $::EOL);
                }
            }
            printf(REPORT "         Sig:", $os);
            printf(REPORT " %s", $versions{$os}->{"defver1"});
            printf(REPORT "  date: %s", $versions{$os}->{"defdate1"});
            print(REPORT $::EOL);
            foreach my $i (2..9) {
                my $defver = "defver" . $i;
                my $defdate = "defdate" . $i;
                if (defined($versions{$os}->{$defver}) or
                    defined($versions{$os}->{$defdate})) {
                    printf(REPORT "         Sig:", $os);
                    if (defined $versions{$os}->{$defver}) {
                        printf(REPORT " %s", $versions{$os}->{$defver});
                    }
                    if (defined $versions{$os}->{$defdate}) {
                        printf(REPORT "  date: %s", $versions{$os}->{$defdate});
                    }
                    print(REPORT $::EOL);
                }
            }
        } # end of LOOP_VER


        printf(REPORT "Company:        %s%s", "", $::EOL);
        printf(REPORT "Address:        %s%s", "", $::EOL);
        printf(REPORT "Phone:          %s%s", "", $::EOL);
        printf(REPORT "Contact:        %s%s", "", $::EOL);
        printf(REPORT "Email:          %s%s", "", $::EOL);
        printf(REPORT "Website:        %s%s", "", $::EOL);


        print(REPORT "Options:",$::EOL);
        LOOP_OPT: foreach $os (sort keys %::osListAll) {
            # check for not tested combinations
            foreach (@::noTestListAll) {
                if ($_ eq "$os ALL $prod") {
                    if ($::debug) {
                        ::log("not tested: $os - ALL - $prod\n", 2);
                    }
                    next LOOP_OPT;
                }
            }
            # check for "NO TEST" status on a specific system
            if (checkNoTestStatus($prod, $os) == 1) {
                # skip this OS because all testbeds have "NO TEST" status
                next LOOP_OPT;
            }

            my $tempLine;
            # handle windows system different (maybe there is just one file
            # for all windows systems (with filename extension "WIN")
            if ($os =~ /^W/i) {
                if ($allWin) {
                    next LOOP_OPT;
                } elsif (-e $infoDir . $::sep . $::configDirNames{"infoOpt"} .
                            $::sep . $::configTestName . ".WIN") {
                    $allWin = 1;
                    $os = "Win";
                }
            }
            my $infoFile = $infoDir . $::sep . $::configDirNames{"infoOpt"} .
                            $::sep . $::configTestName . "." . $os;

            # this key must exist, so initialize it here
            $options{$os} = "";
            if (not -e $infoFile) {
                $options{$os} = "MISSING" . $::EOL;
            } else {
                open(INFO, "< $infoFile") or do {
                    ::log("ERROR: cannot open $infoFile: $!\n", 1);
                    close REPORT;
                    return 1;
                };
                while (defined($tempLine = <INFO>)) {
                    # ignore comment lines (must begin with a '#', ';' or 'REM')
                    if ($tempLine =~ /^(REM|;|\#)/i) { next; }
                    # ignore empty lines
                    if ($tempLine =~ /^\s*$/) { next; }
                    # cut all control chars from end of line
                    $tempLine = ::chommp($tempLine);
                    $options{$os} .= $tempLine . $::EOL;
                }
                close INFO;
            }
            printf(REPORT "    %s: %s", $os, $options{$os});
        } # end of LOOP_OPT

        print(REPORT $::EOL, $::EOL);
    }  #end of foreach LOOP_PROD
    close REPORT;
    return $ret;
}


sub reportProblems {
    my $ret = 0;
    my ($fileHeader, $probHeader, $pslHeader, $specProbHeader);
    # these headertexts are defined in a subroutine because it's
    # easier to read the source this way
    ($fileHeader, $probHeader, $pslHeader, $specProbHeader) = getHeaderProblems();

    # status filenames are set in test configuration
    my $outfile = $::ReportDir . $::sep . $::FileReportProblems;
    open(REPORT, "> $outfile") or do {
        ::log("ERROR: cannot create $outfile: $!\n", 1);
        return 1;
    };
    binmode(REPORT);
    ::log("writing to: $outfile\n", 1);
    print(REPORT $fileHeader, $::EOL, $::EOL, $::EOL);

    print(REPORT $probHeader);

    # create list with postscan information
    ::log("collecting postscanlist information\n", 1);
    print(REPORT $pslHeader, $::EOL);
    $ret = reportProblemsPsl();

    # create list with all problems
    ::log("collecting problems\n", 1);
    print(REPORT $specProbHeader, $::EOL);
    my $prod;
    {
        # write the general problems first
        $prod = "ALL";
        my $genProb = "";
        my $infoDir = $::InfoRoot . $::sep . $prod;
        print(REPORT "General problems:  ",$::EOL);
        my $os = "ALL";
        my $infoFile = $infoDir . $::sep . $::configDirNames{"infoProb"} .
                       $::sep . $::configTestName . "." . $os;
        if (-e $infoFile) {
            open(INFO, "< $infoFile") or do {
                ::log("ERROR: cannot open $infoFile: $!\n", 2);
                $ret = 1;
                next LOOP_PRBLM;
            };
            my $tempLine;
            while (defined($tempLine = <INFO>)) {
                # ignore comment lines (must begin with a '#', ';' or 'REM')
                if ($tempLine =~ /^(REM|;|\#)/i) { next; }
                # ignore empty lines
                if ($tempLine =~ /^\s*$/) { next; }
                # cut all control chars from end of line
                $tempLine = ::chommp($tempLine);
                $genProb .= "      " . $tempLine . $::EOL;
            }
            close INFO;
            print(REPORT $genProb, $::EOL, $::EOL);
        }
    }
    LOOP_PROD2: foreach $prod (sort keys %::prodListAll) {
        my %problems = ();
        my $infoDir = $::InfoRoot.$::sep.$prod;
        print(REPORT $prod,":  ",$::EOL);
        my $os = "ALL";
        # write the problems occuring on all systems first
        $problems{$os} = "";
        my $infoFile = $infoDir . $::sep . $::configDirNames{"infoProb"} .
                       $::sep . $::configTestName . "." . $os;
        if (-e $infoFile) {
            open(INFO, "< $infoFile") or do {
                ::log("ERROR: cannot open $infoFile: $!\n", 2);
                $ret = 1;
                next LOOP_PRBLM;
            };
            my $tempLine;
            while (defined($tempLine = <INFO>)) {
                # ignore comment lines (must begin with a '#', ';' or 'REM')
                if ($tempLine =~ /^(REM|;|\#)/i) { next; }
                # ignore empty lines
                if ($tempLine =~ /^\s*$/) { next; }
                # cut all control chars from end of line
                $tempLine = ::chommp($tempLine);
                $problems{$os} .= "      " . $tempLine . $::EOL;
            }
            close INFO;
            print(REPORT $problems{$os}, $::EOL);
        }
        # and now the problems for each system
        LOOP_PRBLM: foreach $os (sort(keys(%::osListAll))) {
            # check for not tested combinations
            foreach (@::noTestListAll) {
                if ($_ eq "$os ALL $prod") {
                    if ($::debug) {
                        ::log("not tested: $os - ALL - $prod\n", 3);
                    }
                    next LOOP_PRBLM;
                }
            }
            $infoFile = $infoDir . $::sep . $::configDirNames{"infoProb"} .
                        $::sep . $::configTestName . "." . $os;
            if (not -e $infoFile) {
                next LOOP_PRBLM;
            }
            $problems{$os} = "";
            open(INFO, "< $infoFile") or do {
                ::log("ERROR: cannot open $infoFile: $!\n", 2);
                next LOOP_PRBLM;
            };
            my $tempLine;
            while (defined($tempLine = <INFO>)) {
                # ignore comment lines (must begin with a '#', ';' or 'REM')
                if ($tempLine =~ /^(REM|;|\#)/i) { next; }
                # ignore empty lines
                if ($tempLine =~ /^\s*$/) { next; }
                # cut all control chars from end of line
                $tempLine = ::chommp($tempLine);
                $problems{$os} .= "      " . $tempLine . $::EOL;
            }
            close INFO;
            print(REPORT "    ", $os, ":", $::EOL);
            print(REPORT $problems{$os}, $::EOL);
        } # end of LOOP_PRBLM
        print(REPORT $::EOL, $::EOL);
    }  #end of foreach LOOP_PROD2

    close REPORT;
    return 0;
}


sub reportProblemsPsl {
    my $ret = 0;
    my %psl = ();
    LOOP_OS: foreach my $os (sort keys %::osListAll) {
    my $first = 1;
    print(REPORT "  ",$os,":  ");
        LOOP_TB: foreach my $tb (sort keys %::listsAll) {
            my $EvalDir = $::EvalBase . $::sep . $os . $::sep . $tb;
            $psl{$os}->{$tb} = "";

            LOOP_PROD: foreach my $prod (sort keys %::prodListAll) {
                my $statFile = $EvalDir . $::sep . $prod . $::FileExtStatus;
                my $status = ::getStatus($statFile);

                if ($status eq "") {
                    ::log("ERROR getting status for $os - $tb - $prod\n", 2);
                    $ret = 1;
                } elsif ($status =~ /^(-|OK )?S(\d)/i) {
                    my $postscanNr = $2 - 1;
                    if ($postscanNr == 1) {
                        $psl{$os}->{$tb} .= $prod . ",";
                    } elsif ($postscanNr > 1) {
                        $psl{$os}->{$tb} .= $prod . "(" . $postscanNr . "x),";
                    }
                }
            } # end of LOOP_PROD
            # remove last "," from product list
            chop($psl{$os}->{$tb});
            if ($first) {
                $first = 0;
                printf(REPORT "%8s: %s%s", $tb, $psl{$os}->{$tb}, $::EOL);
            } else {
                printf(REPORT "        %8s: %s%s", $tb, $psl{$os}->{$tb}, $::EOL);
            }
        }  #end of foreach LOOP_TB
    } # end of LOOP_OS
    print(REPORT $::EOL, $::EOL);
    return $ret;
}


# contains only some headertext definitions
sub getHeaderProblems {
    my $fileHeader =
'===================================' . $::EOL .
'File 8PROBLMS.TXT:' . $::EOL .
'-----------------------------------' . $::EOL .
'List of problems experienced during ' . $::EOL .
'     VTC test '.$::TestName.':' . $::EOL .
'===================================' . $::EOL .
'Formatted with non-proportional font (Courier)' . $::EOL .
$::EOL .
$::EOL .
'Content of this file:' . $::EOL .
'=====================' . $::EOL .
'    3. Problems of AV products observed during test "'.$::TestName.'"' .
$::EOL .
'       3.1 List of Postscans' . $::EOL .
'       3.2 List of specific problems' . $::EOL;


    my $probHeader =
'3. Problems of AV products observed during test "'.$::TestName.'":'.$::EOL .
'=========================================================='.$::EOL .
$::EOL;


    my $pslHeader =
'3.1 List of Postscans:'.$::EOL .
'----------------------'.$::EOL .
'In several cases, AV/AM products didnot access and check all entries ' .
$::EOL .
'in testbeds (possibly due to the "FF/FN anomaly" as reported in 1.1) ' .
$::EOL .
'or due to crashes or other product misbehaviour (see 3.2). In such cases,' .
$::EOL .
'up to 2 "postscans" were started, whereever possible on the remainder of' .
$::EOL .
'the related testbed.'.$::EOL .
$::EOL .
'The following list summarizes those products where at least 1 postscan ' .
$::EOL .
'was initialised:'.$::EOL;


    my $specProbHeader =
'3.2 List of specific problems:' . $::EOL .
'------------------------------' . $::EOL .
'The following list reports specific problems observed for products' . $::EOL .
'as indicated ("spoon-feeding" means that scanner was restarted on' . $::EOL .
'each subsequent directory when a crash was experienced):' . $::EOL .
$::EOL;


    return ($fileHeader, $probHeader, $pslHeader, $specProbHeader);
}


sub checkNoTestStatus {
    my $prod = $_[0];
    my $os = $_[1];
    # check for "NO TEST" status
    my $noTest = 1;
    foreach my $tb (keys %::listsAll) {
        my $statFile = $::EvalBase . $::sep . $os . $::sep . $tb . 
                       $::sep . $prod . $::FileExtStatus;
        my $status = ::getStatus($statFile);
        $noTest = 0 unless ($status =~ /^NO TEST( .*)?$/i);
    }
    return $noTest;
}

# return a true value
1;