# Tools.pm
# contains useful helper functions for all other modules
#
# v0.83, 2004-01-13
# Copyright (c) Michel Messerschmidt 2002-2004
#
package Avtc::Tools;
use strict;
use Exporter;
use Cwd;
use File::Basename;
use File::Copy;
use File::Path;
use DirHandle;

@Avtc::Tools::ISA = qw(Exporter);
@Avtc::Tools::EXPORT = qw(getStatus updateStatus getDataLine getDataArray
                          winPath unixPath makeDir chommp wc wcSelect
                          appendFile saveFile loadFile deleteTempFiles
                          deleteFiles fileCount archiveFiles mysort
                          indentString getUsername
                          percentRound mover);

#################### general purpose functions #############################

# reads status information from a given file
# and returns the status in uppercase
sub getStatus {
    my $file = $_[0];
    open(STAT, "< $file") or do {
        ::log("ERROR reading $file: $!\n", 2) unless $::quiet;
        # return an empty string to signal an error
        return "";
    };
    binmode(STAT);
    my $line = <STAT>;
    if (not defined($line)) {
        # return an empty string to signal an error
        $line = "";
    } elsif ($line !~ /^(NO .*)|(OK |-)?S\d/i) {
        # return an empty string to signal an error
        $line = "";
    }
    $line = uc(::chommp($line));
    close(STAT);
    return $line;
}


# writes current status to disk
# 1. parameter: status-file, 2. parameter: new status
sub updateStatus {
    my $statFile = $_[0];
    my $status = chommp(uc($_[1]));
    if ($status eq "" or $status !~ /^NO .*|^(OK |-)?S\d(E| )?/i) {
        # don't write illegal status strings
        if (not $::quiet) {
            ::log("ERROR: illegal status '$status' not written\n", 4);
        }
        return 1;
    }
    if ($::debug > 1 ) {
        ::log("DEBUG: writing file $statFile\n", 4);
    }
    my $ok = open(STATFILE, "> $statFile");
    if ($ok) {
        binmode(STATFILE);
        print(STATFILE $status, $::EOL);
    }
    close(STATFILE);
    # return correct values: 0 = normal termination, 1 = error
    # $ok contains return value from open(), which has opposite semantics :-(
    if ($ok) {
        return 0;
    } else {
        ::log("ERROR writing file $statFile\n", 4) unless $::quiet;
        return 1;
    }
}


# read result data from a DAT-file
# this version returns the complete line read from the DAT-file
sub getDataLine {
    my $prod = $_[0];
    my $file = $_[1];
    if (not -e $file) {
        ::log("WARNING: non-existing file $file\n", 2) unless $::quiet;
        return "$prod    !!! DATA FILE NOT FOUND !!!";
    }
    open(DAT, "< $file") or do {
        ::log("ERROR reading $file: $!\n", 2) unless $::quiet;
        return "$prod    !!! DATA FILE NOT FOUND !!!";
    };
    binmode(DAT);
    my $line = <DAT>;
    close(DAT);
    $line = ::chommp($line);
    return $line;
}


# read result data from a DAT-file
# this version returns an array with the single values
sub getDataArray {
    my $file = $_[0];
    if (not -e $file) {
        ::log("WARNING: non-existing file $file\n", 2) unless $::quiet;
        return ("---",-1,-1,-1,-1,-1,-1,-1,-1);
    }
    open(DAT, "< $file") or do {
        ::log("ERROR reading $file: $!\n", 2) unless $::quiet;
        return ("---",-1,-1,-1,-1,-1,-1,-1,-1);
    };
    binmode(DAT);
    my $line = <DAT>;
    close(DAT);
    my @datArray = split(/\s+/, $line);
    return @datArray;
}


# convert a given path from standard-perl/unix- to Dos/Win-convention
# (Backslash instead of slash)
sub winPath {
    my $newPath = $_[0];
    $newPath =~ s-\/-\\-g;
    $newPath =~ s-\\\\-\\-g;
    if ($::debug > 1) {
        ::log("DEBUG: converted $_[0] to $newPath\n", 4);
    }
    return $newPath;
}


# convert a given path from Dos/Win- to Unix-convention
# (Slash instead of backslash)
sub unixPath {
    my $newPath = $_[0];
    $newPath =~ s-\\-\/-g;
    if ($::debug > 1) {
        ::log("DEBUG: converted $_[0] to $newPath\n", 4);
    }
    return $newPath;
}


# check existence of given directory and create it (recursively) if not
# input: 1. directory name to create
#        2. directory access permissions (defaults to 0777)
# returns 0 on error and 1 on success
# This function is also called before the DebugLog file is created,
# so messages can't be written to $::DebugLog
sub makeDir {
    my ($dir, $perm) = @_;
    $perm = 0777 unless defined($perm);
    if (not -e $dir) {
        my $topPath = $dir;
        $topPath =~ s/^(.*)(\\|\/)[^\\\/]+(\\|\/)?$/$1$2/;
#        $topPath .= $::sep;
        if ($::debug > 2) {
            ::log("DEBUG: dir = '".$dir."'  toppath='".$topPath."'\n", 4);
        }
        if ($topPath eq $dir) {
            # we reached the root directory but it doesn't exists
            # (may occur with mapped network drives)
            ::log("ERROR: Failed to create root directory '$dir'\n", 2);
            return 0;
        }
        if (not -e $topPath) {
            my $topOk = makeDir($topPath, $perm);
            if (not $topOk) {
                ::log("ERROR: Failed to create toplevel directory ".
                      "'$topPath'\n", 2);
                return 0;
            }
        }
        if (not -d $topPath) {
            # filename exists but is no directory
            ::log("ERROR: Can't create a directory below a non-directory\n",
                  2);
            return 0;
        }
        my $ok = mkdir($dir, $perm);
        return $ok;
    } elsif (not -d $dir) {
        # filename exists but is no directory
        ::log("ERROR: Filename exists but is no directory\n", 2);
        return 0;
    } else {
        # directory already exists - no need to create it (this is no error)
        return 1;
    }
}


# cut all control characters from begin and end of the input string
# all chars with ASCII code below 32 / 0x20 are considered control chars
sub chommp {
    my $in = $_[0];
    # cut all control characters from end of input string
    while ($in and ord(substr($in, -1, 1)) < 32 ) {
        chop($in);
    }
    # cut all control characters from beginning of input string
    while ($in and ord(substr($in, 0, 1)) < 32 ) {
        $in = substr($in, 1);
    }
    return $in;
}


# no need for an external wc program - here it is
# counts lines and words of input file (char count is not implemented here)
sub wc {
    my $file = $_[0];
    my $lines = 0;
    my $words = 0;
    my $line;
    open(FILE, "< $file") or do {
        return (-1, -1);
    };
    binmode(FILE);
    while (defined($line = <FILE>)) {
        my @currWords = split(/\s+/, chommp($line));
        $words += $#currWords+1;
        $lines++;
    }
    return ($lines, $words);
}


# counts lines and words of input file
# but only when the field given is > 0 in the input file
# (fields are numbered from 0 and separated by spaces)
sub wcSelect {
    my $file = $_[0];
    my $field = $_[1];
    my $lines = 0;
    my $words = 0;
    my $line;

    open(FILE, "< $file") or do {
        return (-1, -1);
    };
    binmode(FILE);
    while (defined($line = <FILE>)) {
        my @currWords = split(/\s+/, ::chommp($line));
        if ($field <= $#currWords and $currWords[$field] > 0) {
            $words += $#currWords+1;
            $lines++;
        }
    }
    return ($lines, $words);
}


# Append a file (2. parameter) to an existing file (1. parameter)
# Insert a linebreak between the files if the 3. parameter is a true value
sub appendFile {
    my ($file, $new, $linebreak) = @_;
    $linebreak = "0" unless defined($linebreak);
    if (not defined($file) or not defined($new)) {
        ::log("ERROR: missing parameter for 'appendFile'\n", 4) unless $::quiet;
        return 0;
    } elsif (not -e $file) {
        ::log("ERROR: missing file '$file' in 'appendFile'\n", 4) unless $::quiet;
        return 0;
    } elsif (not -e $new) {
        ::log("ERROR: missing file '$new' in 'appendFile'\n", 4) unless $::quiet;
        return 0;
        
    }
    if ($::debug > 2) {
        ::log("appending $new to $file\n", 4);
    }
    open(NEW, "< $new") or do {
        return 0;
    };
    binmode(NEW);
    open(FILE, ">> $file") or do {
        close(NEW);
        return 0;
    };
    binmode(FILE);

    if ($linebreak) {
        print FILE $::EOL;
    }
    while (<NEW>) {
        print FILE;
    }
    close(FILE);
    close(NEW);
    return 1;
}


# save (and compress) file to archive
# 1. parameter: file to add to archive
# 2. prameter: directory where the file is
# 3. prameter: directory that contains the archive
sub saveFile {
    my $file = $_[0];
    my $srcDir = $_[1];
    my $archiveDir = $_[2];
    my $extFile = $srcDir . $::sep . $file;
    if ($::configOs eq $::OS_WINDOWS) {
        $extFile = ::winPath($extFile);
    }
    my $extArchive = $archiveDir . $::sep . $::LogArchiveName . "." .
       $::ZipExt;
    if ($::configOs eq $::OS_WINDOWS) {
        $extArchive = ::winPath($extArchive);
    }
    if ($::debug > 1) {
        ::log("saving $file from $srcDir to $archiveDir\n", 4);
    }

    my $extCall = $::ZipProg." ".$::ZipOpt." ".$extArchive." ".$extFile;
    if ($::debug > 1) {
        ::log("calling $extCall\n", 4);
    }
    my $error = system($extCall);
    if ($::debug > 1) {
        ::log("...external program terminated with exitcode $error\n", 4);
    }
    if ($error) {
        ::log("ERROR calling $::ZipProg: exit code=$error, error=$!\n", 4);
        return 0;
    }
    return 1;
}


# load (extract) file from archive
# 1. parameter: file to extract from archive
# 2. prameter: directory that contains the archive
# 3. prameter: directory to extract to
sub loadFile {
    my $file = $_[0];
    my $archiveDir = $_[1];
    my $targetDir = $_[2];

    my $extArchive = $archiveDir . $::sep . $::LogArchiveName . "." .
                     $::ZipExt;
    if ($::configOs eq $::OS_WINDOWS) {
        $extArchive = ::winPath($extArchive);
    }
    if ($::debug > 1) {
        ::log("loading $file from $archiveDir to $targetDir\n", 4);
    }

    # files are always extracted to current directory
    my $oldDir = cwd();
    chdir($targetDir) or do {
        ::log("ERROR: chdir to $targetDir failed: $!\n", 4);
        return 0;
    };
    my $extCall = $::UnzipProg . " " . $::UnzipOpt . " " . $extArchive .
                  " " . $file;
    if ($::debug > 1) {
        ::log("calling $extCall\n", 4);
    }
    my $error = system($extCall);
    if ($::debug > 1) {
        ::log("...external program terminated with exitcode $error\n", 4);
    }
    if ($error) {
        ::log("ERROR calling $::ZipProg: exit code=$error, error=$!\n", 4);
        return 0;
    }
    chdir($oldDir) or do {
        ::log("ERROR: chdir back to $oldDir failed: $!\n", 4);
        return 0;
    };
    return 1;
}


# delete old files in TempDir to avoid broken results,
# but only if we don't want to re-use them
sub deleteTempFiles {
    my $dir = $_[0];
    my $prefix = $_[1];
    my $keepReport = $_[2];

    if ($::debug) {
        ::log("deleting old temporary files\n", 2);
    }
    my @delFiles = ();
    opendir(DIR, $dir) or do {
        ::log("ERROR reading $dir\n", 2);
        ::log("!!! Can not delete old files for $prefix !!!\n", 5);
        return 1;
    };
    my $filename;
    while (defined($filename = readdir(DIR))) {
        if ($filename =~ /^$prefix/i) {
            if ($filename =~ /\.ful$/i and $keepReport) {
                # don't delete report files if 'keepReport' is set
                next;
            }
            push @delFiles, $dir . $::sep . $filename;
            if ($::debug > 2) {
                ::log("deleting " . $dir . $::sep . $filename . "\n", 3);
            }
        } else {
            if ($::debug > 2) {
                ::log("skipping " . $dir . $::sep . $filename . "\n", 4);
            }
        }
    }
    closedir(DIR);
    my $delCount = unlink @delFiles;
    if ($::debug) {
        ::log("$delCount files deleted\n", 2);
    }
    if ($delCount < ($#delFiles + 1)) {
        my $notDel = $#delFiles + 1 - $delCount;
        ::log("ERROR deleting files: $notDel files not deleted\n", 2);
        return 1;
    }
    return 0;
}


# delete files matching $pat (2. parameter) in $dir (1.parameter)
# if 3. parameter is set to a true value, _all_ files in all
# subdirectories will be deleted too
# ATTENTION: the file pattern is case independent
#            (but that may change in future versions)
sub deleteFiles {
    my ($dir, $pat, $rec, $topDir) = @_;
    # default is delete all files
    $pat = ".*" unless defined($pat);
    $rec = 0 unless defined($rec);
    $topDir = $dir unless defined($topDir);
    my $ret = 0;
    # remember current directory
    my $prevDir = cwd();
    if ($::debug > 2) {
        ::log("DEBUG: dir = '$dir'  -  pattern = '$pat'  -  recursive = ".
              "'$rec'\n", 3);
    }

    if (not -e $dir) {
        ::log("ERROR: trying to delete in non-existing directory '$dir'\n", 2);
        return 1;
    }
    if (not -d $dir) {
        ::log("ERROR: '$dir' is not a directory\n", 2);
        return 1;
    }

    chdir($dir) or do {
        ::log("ERROR: chdir $dir failed:  $!\n", 1);
        ::log("Deletion of files in '$dir' failed !\n", 4);
        return 1;
    };
    if ($::debug > 1) {
        ::log("Current dir set to ". cwd() ."\n", 1);
        ::log("deleting files in $dir\n", 2);
    }
    # get a list of all files in this directory
    my $dirH = new DirHandle "." or do {
        ::log("ERROR: reading '$dir'\n", 3);
        # change back to previous directory
        chdir($prevDir);
        return 1;
    };
    # create a list of all files that should be deleted
    my @delFiles = ();
    my $file;
    while (defined($file = readdir($dirH))) {
        if ($file =~ /^\.\.?$/) {
            # skip self-referential and topdirectory entries
            next;
        }
        if (-d $file) {
            # found a directory
            if ($rec) {
                $ret = deleteFiles($file, '.*', 1, $topDir);
            }
        } elsif ($file =~ /$pat/i) {
            push @delFiles, $file;
            if ($::debug > 2) {
                ::log("deleting $file\n", 3);
            }
        } else {
            if ($::debug > 2) {
                ::log("skipping $file\n", 4);
            }
        }
    }
    $dirH->close;

    my $delCount = unlink @delFiles;
    if ($::debug > 1) {
        ::log("$delCount files deleted\n", 2);
    }

    # change back to previous directory
    chdir($prevDir);

    if ($delCount < ($#delFiles + 1)) {
        my $failed = $#delFiles+1-$delCount;
        ::log("ERROR deleting files: $failed files not deleted: $!\n", 2);
        $ret = 1;
    } else {
        # try to delete the directory itself (but only subdirectories not the
        # directory this functions was called with)
        if ((fileCount($dir) == 0) and ($dir ne $topDir)) {
            my $deleted = rmdir $dir;
            if (not $deleted) {
                ::log("ERROR deleting directory '$dir': $!\n", 2);
                $ret = 1;
            }
        }
    }

    return $ret;
}

# Takes a directory name and returns the number of files (incl. directories)
# in it. Returns -1 on errors and 0 for empty directories.
sub fileCount {
    my $dir = $_[0];
    if (not -e $dir) {
        ::log("ERROR: directory '$dir' doesn't exist\n", 2);
        return -1;
    }
    if (not -d $dir) {
        ::log("ERROR: '$dir' is not a directory\n", 2);
        return -1;
    }
    $dir .= $::sep unless ($dir =~ /$::sep$/);
    my $dirH = new DirHandle "$dir" or do {
        ::log("ERROR: reading '$dir'\n", 3);
        return -1;
    };
    my $f;
    my $count = 0;
    while (defined($f = readdir($dirH))) {
        if ($f =~ /^\.\.?$/) {
            # skip self-referential and topdirectory entries
            next;
        }
        $count++;
    }
    $dirH->close;
    return $count;
}

# save all data created during evaluation to one archive
# 1. parameter: the archive name,
# 2. parameter: dir to get files from
# 3. parameter: dir to save archive in
sub archiveFiles {
    my $prefix = $_[0];
    my $fromDir = $_[1];
    my $toDir = $_[2];
    my $archive =  $prefix . "." . $::ZipExt;
    my $wildcard = "*.*";
    if ($::configOs eq $::OS_UNIX) {
        $wildcard = "*";
    }

    my $oldDir = cwd();
    chdir($fromDir) or do {
        ::log("ERROR: chdir to $fromDir failed: $!\n", 4);
        return 0;
    };

    if ($::debug > 1) {
        ::log("compressing evaluation files\n", 4);
    }
    my $extCall = $::ZipProg . " " . $::ZipOpt . " " . $archive . " " .
                  $prefix . $wildcard;
    if ($::debug > 1) {
        ::log("calling $extCall\n", 4);
    }
    my $error = system($extCall);
    if ($::debug > 1) {
        ::log("...external program terminated with exitcode $error\n", 4);
    }
    if ($error) {
        ::log("ERROR calling $::ZipProg: exit code=$error, error=$!\n", 4);
        return 0;
    }

    my $ok = copy($fromDir . $::sep . $archive, $toDir . $::sep . $archive);
    # no error-checking here, only needed if diskspace is low on TempDir
    unlink($fromDir. $::sep . $archive);

    chdir($oldDir) or do {
        ::log("ERROR: chdir back to $oldDir failed: $!\n", 4);
        return 0;
    };

    return $ok;
}


# sort file contents and write to another file
# input and output file may be the same file
sub mysort {
    my $in = $_[0];
    my $out = $_[1];
    my $delIdent = $_[2]; # should identical lines be deleted
    # should an external program be used ?
    if ($::extSort) {
        my $extTemp = $::TempBase;
        if ($::configOs eq $::OS_WINDOWS) {
            $extTemp = ::winPath($extTemp);
            $in = ::winPath($in);
            $out = ::winPath($out);
        }
        # "-T" sets dir for temporary files, "-o" specifies output file
        # old scripts called sort with option "-u" to delete identical
        # lines - that is now considered an error
        my $extCall = $::SortProg ." -T". $extTemp;
        if ($delIdent) {
            $extCall .= " -u -o ". $out ." ". $in;
        } else {
            $extCall .= " -o ". $out ." ". $in;
        }
        if ($::debug > 1) {
            ::log("calling $extCall\n", 4);
        }
        my $error = system($extCall);
        if ($error) {
            ::log("ERROR calling sort: exit code=$error, error=$!\n", 4);
            return 0;
        }
        if ($::debug > 1) {
            ::log("...external program terminated with exitcode $error\n", 4);
        }
    } else {
        open(SORTIN, "< $in") or do {
            ::log("ERROR opening $in: $!\n", 4);
            ::log("ERROR sorting $in\n", 4);
            return 0;
        };
        binmode(SORTIN);
        my @tempIn = ();
        my $i;
        while(defined($i = <SORTIN>)) {
            $i = chommp($i);
            push @tempIn, $i;
        }
        close(SORTIN);

        my @tempOut = sort @tempIn;
        undef @tempIn;

        open(SORTOUT, "> $out") or do {
            ::log("ERROR writing $out: $!\n", 4);
            ::log("ERROR sorting $in\n", 4);
            return 0;
        };
        binmode(SORTOUT);
        my $j;
        my $last = "";
        foreach $j (@tempOut) {
            if ($delIdent and ($last eq $j)) {
                # skip identical lines
                next;
            }
            print(SORTOUT $j, $::EOL);
            $last = $j;
        }
        close(SORTOUT);
    }
    return 1;
}


# This function wraps all lines to 78 characters and indents all wrapped
# lines with whitespace up to a given substring of the line
# Input: 1. string to operate on
#        2. substring that marks the end of the indent (':' for example)
sub indentString {
#    my $in = $_[0];
#    my $headEnd = $_[1];
#    my $lineEnd = $_[2];
    my ($in, $headEnd, $lineEnd) = @_;
    $lineEnd = 78 unless ($lineEnd > 0);

    my $out = $in;
    my $headindex = 0;
    my $head = "";
    my $eolen = length($::EOL);

    $in =~ /\Q$headEnd\E/g;
    $headindex = pos($in);
    if (not defined $headindex) { $headindex = 0; }
    $head = " "x$headindex;
    if ($::debug > 2) {
        ::log("\n");
        ::log("DEBUG:    line end = $lineEnd\n");
        ::log("DEBUG:    initial headindex=$headindex\n");
    }

    my $endpos = 0;
    my $startpos = 0;
    while ($endpos < length($out)) {
        # get position of next newline
        # option s treats string as single line (to match newlines),
        # option g ("global" search) searches ALL pattern occurences in
        # subsequent regex calls
        $out =~ /\n/gs;
        $endpos = pos($out);
        if (not defined($endpos) or $endpos < 1 or $endpos > length($out)) {
            $endpos = length($out);
        }
        if ($::debug > 2) {
            ::log("DEBUG:    found new line end = $endpos\n");
            ::log("DEBUG:    startpos=".$startpos."  -  endpos=".$endpos.
                  "  -  length=".$endpos-$startpos."\n");
            ::log("DEBUG:    line='".
                  substr($out, $startpos, $endpos-$startpos) . "'\n");
        }

        while ($endpos - $startpos > $lineEnd) {
            my $newpos = $startpos + $lineEnd;
            while (substr($out, $newpos, 1) ne " ") { $newpos--; }

            my $tempStr = substr($out, $startpos, $newpos-$startpos);
            if ($tempStr =~ /\Q$headEnd\E/g) {
                $headindex = pos($tempStr);
                $head = " "x$headindex;
            }
            if ($::debug > 2) {
                ::log("DEBUG:    head='".$head."'  -  headindex=".$headindex.
                      "\n");
                ::log("DEBUG:    newpos=".$newpos."\n");
                ::log("DEBUG:    shortline='".
                      substr($out, $startpos, $newpos-$startpos) . "'\n");
            }
            substr($out, $newpos, 1) = substr($out, $newpos, 1) . $::EOL .
                                       $head;
            pos($out) = $newpos + 1 + $eolen;
            # newpos is the last char in the shorter line, newpos+1 is the line feed,
            # newpos+1+$eolen is the first char in the new line
            $startpos = $newpos + 1 + $eolen;
            $endpos += length($head) + $eolen;
            if ($::debug > 2) {
                ::log("DEBUG:    newstart=".$startpos."  -  newend=".$endpos.
                      "  -  length=". $endpos-$startpos ."\n");
                ::log("DEBUG:    newline='".
                      substr($out, $startpos, $endpos-$startpos) . "'\n");
            }
        }
        # endpos is the last char in the old line, endpos+1 is the line feed,
        # endpos+2 is the first char in the next line
        $startpos = $endpos;
    }

    return $out;
}


sub getUsername {
    my $name = "";
    if (not $::gui) {
        ::log("Later it may be important to know who made this status ".
              "change.\n");
        my $nameOK = 0;
        while (not $nameOK) {
            ::log("Enter your name: ");
            $name = <STDIN>;
            chomp($name);
            if ($name eq "") {
                ::log("ERROR: An empty name is not acceptable\n");
            } else {
                $nameOK = 1;
            }
        }
    } else {
        # and now the GUI version
        my(@ll) = qw/-anchor e -justify right/;
        my(@el) = qw/-justify left -relief sunken/;
        my(@rl) = qw/-anchor w -justify left/;
        my $msg = "Later it may be important to know who made this status ".
                  "change.\n\n";
        my $errmsg = "\n\n\n";
        my $nameOK = 0;
        while (not $nameOK) {
            my $dlg = $::mainWin->DialogBox(-title => "Enter your name",
                                       -buttons => ["Ok"]
#                                       , -default_button => "Ok"
                                       );
            my $l1 = $dlg->add("Label", @rl, -text => $msg)
                ->pack();
            my $l2 = $dlg->add("Label", @ll, -text => "Enter your name:")
                ->pack();
            my $w2 = $dlg->add("Entry", @el, -width => 30,
                                        -textvariable => \$name)
                ->pack();
            my $l3 = $dlg->add("Label", @rl, -text => $errmsg)
                ->pack();

            # show dialog
            $dlg->Show;
            if ($name ne "") {
                $nameOK = 1;
            } else {
                $errmsg = "\n\nERROR: An empty name is not acceptable\n";
            }
        }
    }
    return $name;
}


###
### functions that are currently not used but
###  - should replace older code in various locations
###  - may be useful later
###
# round a percent value to 100.~ or 0.~
# parameters: 1. percent value to round
#             2. numerical value, the percent value is belonging to
#             3. numerical value that equals 100.0%
sub percentRound {
    my $perc = sprintf("%5.1f", $_[0]);
    my $nr = $_[1];
    my $total = $_[2];
    if ($perc eq "100.0" and $nr < $total) {
        $perc = "100.~";
    } elsif ($perc eq "  0.0" and $nr > 0) {
        $perc = "  0.~";
    }
    return $perc;
}


# move files and directories recursively
# parameters: 1: source file/directory
#             2: target file/directory
#             3: overwrite-flag: 0 = keep old files,  1 = overwrite files
sub mover {
    my ($src, $dest, $over) = @_;
    if (not -e $src) {
        ::log("ERROR: trying to move non-existing file\n", 2);
        return 0;
    }
    if (-e $dest and not -d $dest) {
        if ($over) {
### not implemented yet
        } else {
            ::log("ERROR: move destination file already exists\n", 2);
            return 0;
        }
    }
    if (not -d $src) {
        # move a single file
        if (-d $dest) {
            my $f = $src;
            $f =~ s/^.*(\\|\/)([^\\\/]+)$/$2/;
            $dest .= $::sep . $f;
            if ($::debug > 1) {
                ::log("DEBUG: new dest = ',".$dest."'\n", 4);
            }
            if (-e $dest) {
                if ($over) {

                } else {
                    ::log("ERROR: move destination already exists\n", 2);
                    return 0;
                }
            }
        }
        move($src, $dest) or do {
            ::log("ERROR moving '".$src."' to '".$dest."': $!\n", 2);
        };
    } else {
        if (not -e $dest) {
            makeDir($dest . $::sep) or do {
                ::log("ERROR: Failed to create destination directory\n", 2);
                return 0;
            };
        }
        my $ok = 1;
        my $dirH = new DirHandle "$src.$::sep" or do {
            ::log("ERROR reading ".$src.$::sep.": $!\n", 3);
            return 0;
        };
        my $f;
        while (defined($f = readdir($dirH))) {
            if ($f =~ /^\.\.?$/i) {
                # skip self-referential and topdirectory entries
                next;
            }
            if (-d $f) {
                $ok = mover($src . $::sep . $f, $dest . $::sep . $f, $over);
            } else {
                my $destFile = $dest . $::sep . $f;
                if (-e $destFile) {
                    if ($over) {
### not implemented yet
                    } else {
                        ::log("ERROR: move destination file already exists\n",
                              2);
                        return 0;
                    }
                }
                $ok = mover($src . $::sep . $f, $destFile, $over);
            }
        }
        $dirH->close;
        return $ok;
    }
}


# return a true value
1;