#!/usr/bin/perl -w
#
# avtctest.pl - the aVTC test evaluation tool
#
# Copyright (C) 2002-2004  Michel Messerschmidt
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
use strict;
use File::Copy;
use File::Basename;
use Cwd;
use Avtc::Tools;      # contains useful helper functions for all modules
use Avtc::Make;       # implements 'new' and 'update' commands
use Avtc::Create;     # 'create' command
use Avtc::Eval;       # 'eval', 'reeval' and 'archive' commands
use Avtc::Report;     # 'result' and 'status' commands
use Avtc::Rpt_info;   # 'report info' and 'report problems' commands
use Avtc::Rpt_comp;   # 'report compare' and 'check compare' command
use Avtc::Ranking;    # 'report rank' command (only for heureka tests)
use Avtc::Status;     # 'status' and all commands that change the status of
                      # a reportfile
use Avtc::Export;     # 'report logs' and 'report missed' commands
use Avtc::Othercmd;   # all other commands ('show', ...)
use Avtc::GUI;        # implements graphical user interface

# get start time as soon as possible
my $startTime = localtime(time);
# save current directory for later restore
my $prevDir = cwd();

### Set constant expressions ###
$::prog = basename($0);
$::version = "v2.62, 2004-07-09 written by Michel Messerschmidt";
# Constant values to identify the different operating systems
$::OS_WINDOWS="windows";
$::OS_UNIX="unix";
# Path separator characters for the different operating systems
$::sep_WINDOWS = "\\";
$::sep_UNIX = "/";
# Define end-of-line characters independant from any operating system or
# perl conventions.
# (It should always be CR+LF (windows syntax) for tests at the aVTC)
$::EOL = "\x0d\x0a";
# This string is used by scanner-keys as virusname to signal
# "unreliable identification"
$::VirusNameUI = "NOT_IDENTIFIED";
# This string describes all valid sample filenames as a regular expression
# and is used by scanner-keysto decide if a given filename is a valid sample
$::samplenameRE = '_[A-Z0-9]{4}_[0-9]{1}\.[A-Z0-9_]{1,3}';
# Maximum number of scans to accept for one reportfile (1.scan + postscans)
$::maxScan = 3;

### These global variables are set accordingly to the current system ###
# This contains the operating system that we're currently running on
$::configOs = "";
# Path separator character for this operating system
$::sep = "";
# Path to global configuration file
# This should be the ONLY hardcoded file in the this program !!!
# (another configuration file can be specified on the commandline)
$::configGlobalFile = "";
# $^O contains the name of the system that the perl executable was compiled on
if ($^O =~ /MSWin/i) {
    $::configOs = $::OS_WINDOWS;
    $::sep = $::sep_WINDOWS;
    $::configGlobalFile = "P:\\auswert\\scripte\\test.cfg";
#    $::ScriptDir = "P:\\auswert\\scripte";
} else {
    # assume a system following unix conventions as default
    $::configOs = $::OS_UNIX;
    $::sep = $::sep_UNIX;
    $::configGlobalFile = "/server/p/auswert/scripte/test.cfg";
#    $::ScriptDir = "/server/p/auswert/scripte";
}

### Define some global configuration variables and their default values ###
# Use perls internal sort function (0) or an external sort program (1) ?
# The internal function is much faster but needs lots of memory, while the
# external sort program uses temporary files on disk (in $::TempRoot)
# (these defaults may be changed by the test-specific configuration file)
$::extSort = "0";
$::SortProg = "";
# Default debug level (currently used levels: 0 - 3)
$::debug = 0;

# Set list of all known commands
%::cmds = ();
defineCommands();
# Default command is to start the graphical user interface
$::command = "";

# Set list of all known status codes
%::statusCodes = ();
defineStatusCodes();
# Set default status code for 'reset'
$::resetStatus = "-S1";

# Set global variables containing different filename parts (extensions, ...)
defineFilenameParts();
# ...and help texts
defineHelp();

# Some global boolean flags (0 = no, 1 = yes):
# Use graphical user interface
$::gui = 0;
# This is set to produce no output on STDOUT, only a return value
$::quiet = 0;
# If this is set, wait for a keypress after each 24 lines of output on STDOUT
$::pause = 0;
# Some flags to avoid loading of (not always needed) configuration files
$::needsPack = 0;
# These flags are only set to a true value (1) if the corresponding
# configuration files were loaded without errors
$::loadedGlobalConfig = 0;
$::loadedTestConfig = 0;
$::loadedTestLists = 0;
$::loadedSum = 0;
$::loadedPack = 0;
$::loadedPackCreate = 0;
# This selects whether available contents of TempDir will be deleted or reused
$::skipReload = 0;
# If true, all specified logfiles with status S[1-3]E will be evaluated again
$::reEval = 0;
# This flag selects, if products will be evaluated or just copied to LogDir
$::archiveOnly = 0;

# initialize parameter arrays
my @optionOs = ();
my @optionTb = ();
my @optionProd = ();
my @optionPack = ();

### Initialize some global configuration variables ###
# Path to test specific configuration file
# (this is read from the global configuration file)
$::configTestFile = "";
# Name of test to process (read from commandline or global configuration file)
$::configTestName = "";
# Root directory for temporary files/directories
# (this is read from the global configuration file)
$::TempRoot = "";
# How to interpret/convert file paths in input and output files
# (this is set in the test-specific configuration file)
$::configSyntax = "";
# Path separator character used in scan reports (depends on $configSyntax)
$::rps = "";
# Initialize some variables used to hold configuration info
$::configRoot = "";    # root directory of the whole evaluation
%::configRoot = ();    # root directories for all supported systems
$::configDebugFile = ""; # filename for debug/log output
%::configDirNames = ();  # list of directory names used in evaluation system
%::configAllTests = ();  # list of all known tests and configuration files
%::osList = ();        # list of selected operating systems
%::osListAll = ();     # list of all operating systems
%::prodList = ();      # list of selected products
%::prodListAll = ();   # list of all products
%::lists = ();         # contains selected testbeds data (from LISTS dir)
%::listsAll = ();      # contains all testbeds data (from LISTS dir)
%::packList = ();      # list of selected compression programs
%::packListAll = ();   # list of all compression programs
@::noTestListAll = (); # list of all os-testbed-product combinations that
                       # are not tested
%::data = ();          # used to store dynamic data that need not be saved
                       # in a file

# save argument array for debug output
my @DEBUGARG = @ARGV;

while (defined $ARGV[0] and $ARGV[0] =~ s/^-|^\///) {
    if ($ARGV[0] =~ /^(v|-?version)$/i) {
        print("$::prog $::version\n",
	      "Copyright (C) 2002-2004 Michel Messerschmidt\n",
	      "This program may be used under the terms of the ",
	      "GNU General Public License.\nSee the file COPYING for ",
              "details.\n\n");
        exit 0;
    } elsif ($ARGV[0] =~ /^(h|\?)$/i) {
        displayUsage("short");
        exit 0;
    } elsif ($ARGV[0] =~ /^-?help$/i) {
        displayUsage("");
        exit 0;
    } elsif ($ARGV[0] =~ /^(q|-?quiet)$/i) {
        $::quiet = 1;
        shift;
    } elsif ($ARGV[0] =~ /^(p|-?pause)$/i) {
        $::pause = 1;
        shift;
    } elsif ($ARGV[0] =~ /^-?debug$/i) {
        $::debug++;
        shift;
    } elsif ($ARGV[0] =~ /^(d+)$/i) {
        $::debug += length($1);
        shift;
    } elsif ($ARGV[0] =~ /^-?c(onfig)?(=.*)?$/i) {
        # may be given as '-c=file' or '-c file'
        if (defined($2)) {
            $::configGlobalFile = substr($2, 1);
        } else {
            shift;
            # next argument is the filename
            # this may even be true, when something liken '-c=file' is given,
            # because some command interpreters automatically split this in
            # two arguments '-c' and 'file' (the '=' is discarded)
            $::configGlobalFile = $ARGV[0];
        }
        shift;
    } elsif ($ARGV[0] =~ /^(n|-?test)(=.*)?$/i) {
        # may be given as '-n=name' or '-n name'
        if (defined($2)) {
            $::configTestName = substr($2, 1);
        } else {
            shift;
            $::configTestName = $ARGV[0];
        }
        shift;
    } elsif ($ARGV[0] =~ /^-?t(emp)?(=.*)?$/i) {
        # may be given as '-t=name' or '-t name'
        if (defined($2)) {
            $::TempRoot = substr($2, 1);
        } else {
            shift;
            $::TempRoot = $ARGV[0];
        }
        shift;
    } elsif ($ARGV[0] =~ /^-?testconfig(=.*)?$/i) {
        # may be given as '-testconfig=name' or '-testconfig name'
        if (defined($1)) {
            $::configTestFile = substr($1, 1);
        } else {
            shift;
            $::configTestFile = $ARGV[0];
        }
        shift;
    } elsif ($ARGV[0] =~ /^(s|-?skipReload)$/i) {
        $::skipReload = 1;
        shift;
    } else {
        print("ERROR: unknown option\n\n",
              "Type '$::prog help' for more information\n\n");
        if ($::debug) {
            my $i=0;
            while ($i <=$#DEBUGARG) {
                print("DEBUG: ARGV[",$i,"] = ",$DEBUGARG[$i],"\n");
                $i++;
            }
        }
        exit 1;
    }
}
# quiet overrides all debug levels
if ($::quiet) {
    $::debug = 0;
}
# handle command parameter
if (not defined $ARGV[0]) {
    # start the gui if no parameters are given
    $::command = "gui";
}
elsif ($ARGV[0] =~ /^(help|new|update|archive|set|reset|status|extract|clean|cleanall|gui)$/i) {
    $::command = lc($ARGV[0]);
    shift;
} elsif ($ARGV[0] =~ /^(eval|reeval|result|resultcsv)$/i) {
    $::command = lc($ARGV[0]);
    $::needsSum = 1;
    shift;
} elsif ($ARGV[0] =~ /^create$/i) {
    $::command = lc($ARGV[0]);
    shift;
} elsif ($ARGV[0] =~ /^show$/i) {
    $::command = lc($ARGV[0]) . " ";
    shift;
    if ($#ARGV < 0) {
        print("ERROR: you must specify a list or data source to show ",
          "(eg. 'show tests')\n\n");
        exit 1;
    }
    if ($ARGV[0] =~ /^(config|tests|products|testbeds|systems|pack|status)$/i) {
        $::command .= lc($ARGV[0]);
    } elsif ($ARGV[0] =~ /^(packcmd|dump)$/i) {
        $::command .= lc($ARGV[0]);
    } else {
        print("ERROR: unknown data source to show\n\n");
        exit 1;
    }
    shift;
} elsif ($ARGV[0] =~ /^report$/i) {
    $::command = lc($ARGV[0]) . " ";
    shift;
    if ($#ARGV < 0) {
        print("ERROR: you must specify a report part (eg. 'info')\n\n");
        exit 1;
    }
    if ($ARGV[0] =~ /^(info|problems|compare|logs|missed)$/i) {
        $::command .= lc($ARGV[0]);
    } elsif ($ARGV[0] =~ /^rank$/i) {
        $::command .= lc($ARGV[0]);
        $::needsSum = 1;
    } else {
        print("ERROR: unknown report part\n\n");
        exit 1;
    }
    shift;
} elsif ($ARGV[0] =~ /^check$/i) {
    $::command = lc($ARGV[0]) . " ";
    shift;
    if ($#ARGV < 0) {
        print("ERROR: you must specify a check\n\n");
        exit 1;
    }
    if ($ARGV[0] =~ /^compare$/i) {
        $::command .= lc($ARGV[0]);
        $::needsSum = 1;
    } else {
        print("ERROR: unknown check\n\n");
        exit 1;
    }
    shift;
} else {
    print("ERROR: unknown command\n\n",
          "Type '$::prog help' for more information\n\n");
    if ($::debug) {
        my $i=0;
        while ($i <=$#DEBUGARG) {
            print("DEBUG: ARGV[",$i,"] = ",$DEBUGARG[$i],"\n");
            $i++;
        }
    }
    exit 1;
}

# get the additional status parameter if 'reset' command is selected
if ($::command eq "set") {
    if (defined $ARGV[0] and $ARGV[0] =~ /^ok|postscan$/i) {
        $::command .= " ". lc($ARGV[0]);
        shift;
        if ($::debug and not $::quiet) {
            print("DEBUG: found correct (re)set option. Command is now '",
                  $::command,"'\n");
        }
    } else {
        print("ERROR: unknown status: use either 'ok' or 'postscan'\n\n");
        exit 1;
    }
}
# get the additional status parameter if 'reset' command is selected
if ($::command eq "reset") {
    if (defined $ARGV[0]) {
        if ($ARGV[0] =~ /^ok|postscan$/i) {
            $::command .= " ". lc($ARGV[0]);
            shift;
            if ($::debug and not $::quiet) {
                print("DEBUG: found correct (re)set option. Command is now '",
                      $::command,"'\n");
            }
        } else {
            $::resetStatus = uc($ARGV[0]);
            shift;
            if ($::debug and not $::quiet) {
                print("DEBUG: found reset option: '",$::resetStatus,"'\n");
            }
        }
    }
    # else: default status is set some lines above
}
# get os, tb and product parameter or use 'ALL' if not given.
# 'update', 'create', 'result compare' and 'check compare' are the
# only commands that don't take an OS parameter
if (defined $ARGV[0] and ($::command ne "update") and ($::command ne "create")
    and ($::command !~ /compare$/)) {
    @optionOs = split(",", uc($ARGV[0]));
    shift;
} else {
    @optionOs = ("ALL");
}
if (defined $ARGV[0]) {
    @optionTb = split(",", uc($ARGV[0]));
    shift;
} else {
    @optionTb = ("ALL");
}
if (defined $ARGV[0]) {
    # 3. parameter is the product for all command except 'create'
    # which takes a compression program instead of the product parameter
    if ($::command eq "create") {
        @optionPack = split(",", uc($ARGV[0]));
        @optionProd = ("ALL");
    } else {
        @optionProd = split(",", uc($ARGV[0]));
        @optionPack = ("ALL");
    }
    shift;
} else {
    @optionProd = ("ALL");
    @optionPack = ("ALL");
}

if ($#optionOs > 0) {
    my @newOpt = ();
    foreach my $i (@optionOs) {
        if ($i ne "ALL") {
            push @newOpt, $i;
        }
    }
    @optionOs = @newOpt;
}
if ($#optionTb > 0) {
    my @newOpt = ();
    foreach my $i (@optionTb) {
        if ($i ne "ALL") {
            push @newOpt, $i;
        }
    }
    @optionTb = @newOpt;
}
if ($#optionProd > 0) {
    my @newOpt = ();
    foreach my $i (@optionProd) {
        if ($i ne "ALL") {
            push @newOpt, $i;
        }
    }
    @optionProd = @newOpt;
}
if ($#optionPack > 0) {
    my @newOpt = ();
    foreach my $i (@optionPack) {
        if ($i ne "ALL") {
            push @newOpt, $i;
        }
    }
    @optionPack = @newOpt;
}


if ($::debug > 2 and not $::quiet) {
    my $i=0;
    while ($i <=$#DEBUGARG) {
        print("DEBUG: ARGV[",$i,"] = ",$DEBUGARG[$i],"\n");
        $i++;
    }
}

# commands that don't need the configuration files are executed here,
# all others after loading the configuration data
if ($::cmds{$::command}{"needsNoGlobal"}) {
    my $ret = execCmd();
    exit $ret;
}

# read global configuration file
getGlobalConfig() or die "\nERROR reading global configuration\n\n";
# check existance of TempDir and create it if nessecary (required for logfile)
::makeDir($::TempRoot) or die "ERROR: creating directory $::TempRoot\n";
# create log file
my $logFile = $::TempRoot . $::sep . $::configDebugFile;
$::DebugFile = new IO::File;
$::DebugFile->open("> $logFile") or die
    "ERROR: can't create $logFile: $!\n\n";
autoflush $::DebugFile 1;
# set flag to signal that global configuration and logfile can be used
$::loadedGlobalConfig = 1;
if (not $::quiet) {
    ::log($::prog." started on ".$startTime."\n", 0);
}

# get name of test specific configuration file
# only use default configuration file if not given on cmdline
$::configTestFile = $::configAllTests{$::configTestName}
    unless $::configTestFile;


# commands that depend only on global config file are executed here,
# all others after loading the test-specific configuration data
if ($::cmds{$::command}{"needsNoConfig"}) {
    my $ret = execCmd();
    exitPrg($ret);
}


# now read the test-specific configuration file
# abort if testname is unknown (= there was no section in global config)
readTestConfig() or do {
    ::log("\nERROR reading test configuration\n\n", 0);
    exitPrg(1);
};


if ($::debug > 3) {
    ::log("Perl Library path:\n", 0, "DEBUG: ");
    foreach (@INC) {
        ::log("  $_\n", 1, "DEBUG: ");
    }
    ::log("commandline is: $::prog @DEBUGARG \n", 0, "DEBUG: ");
    ::log("List of all test names and configuration files:\n", 0, "DEBUG: ");
    foreach my $key (keys %::configAllTests) {
        ::log("name: ".$key."  -  file: ".$::configAllTests{$key}."\n", 2,
              "DEBUG: ");
    }
    ::log("Default test: '$::configTestName'\n", 1, "DEBUG: ");
    ::log("Default config: '$::configTestFile'\n", 1, "DEBUG: ");
}


# get lists of all systems/testbeds/products/...
getTestLists() or do {
    ::log("\nERROR reading test configuration lists\n\n", 0);
    exitPrg(1);
};


# now we have all information to execute any other command
if (not $::quiet) {
    ::log("Working on test: ".$::configTestName."\n", 0);
}
# execute command
my $ret = execCmd();
exitPrg($ret);

######################### end of main program ###############################



# write messages to all available output devices (STDOUT, logfile, logwindow)
# parameters: $out = message, $i = indentation level, $pre = line prefix
sub log {
    my ($out, $i, $pre) = @_;
    $i = 0 unless defined($i);
    $pre = "" unless defined($pre);
    $out = $pre . "  "x$i . $out;
    print($out);
    if (defined($::DebugFile)) {
        print($::DebugFile $out);
    }
    if ($::gui) {
        $::outMsg->insert("end", $out);
        $::outMsg->see("end");
        $::outMsg->update;
    }
}


# execute the selected command, if possible
sub execCmd {
    if ($::cmds{$::command}{"needsSum"} and not $::loadedSum) {
        ::log("ERROR executing command:\n".
              "      Summary file is missing but required for '$::command'\n".
              "      You must run the 'update' command to create them.\n");
        return 1;
    }
    if ($::cmds{$::command}{"needsPackCreate"} and not $::loadedPackCreate) {
        ::log("ERROR executing command:\n".
              "      Missing configuration files for '$::command'\n");
        return 1;
    }
    # some commands require confirmation
    if (not $::gui and $::cmds{$::command}{"ask"}) {
        print("Really execute '$::command' (y/n) ?   ");
        my $input = <STDIN>;
        chomp($input);
        if (lc($input) ne "y")  {
            ::log("Execution of command '$::command' canceled.\n", 0);
            return 1;
        }
    }
    my $ret;
    if (exists $::cmds{$::command}{"opt"}) {
        ::log("Executing command '".$::command."' with options '".
              ${$::cmds{$::command}{"opt"}}."' (" .
              $::cmds{$::command}{"desc"} . ")\n");
        $ret = &{$::cmds{$::command}{"exec"}}(${$::cmds{$::command}{"opt"}});
    } else {
        ::log("Executing command '".$::command. "' (" .
              $::cmds{$::command}{"desc"} . ")\n");
        $ret = &{$::cmds{$::command}{"exec"}}();
    }
    if ($ret) {
        ::log("ERROR executing command\n", 0);
    }
    ::log("...done\n", 0);
    return $ret;
}


# function to do some cleanup before program exit
sub exitPrg {
    my $ret = $_[0];
    $ret = 0 unless defined($ret);
    # change back to the directory we started in
    my $nowDir = cwd();
    if ($prevDir ne $nowDir) {
        chdir($prevDir);
        if ($::debug > 2) {
            print($::DebugFile "chdir back to $prevDir\n");
        }
    }
    # record end time
    my $endTime = localtime(time);
    if (not $::quiet) {
        print("\n...done on ",$endTime,"\n");
    }
    print($::DebugFile "\n...done on ",$endTime,"\n");
    $::DebugFile->close;
    exit $ret;
}


# read the global configuration file
sub getGlobalConfig {
    if (not -e $::configGlobalFile) {
        print("ERROR: global configuration file not found: ",
              $::configGlobalFile,"\n\n");
        return 0;
    }
    open(CONFIGGLOBAL, "< $::configGlobalFile") or do {
        print("ERROR opening $::configGlobalFile: $!\n\n");
        return 0;
    };
    binmode(CONFIGGLOBAL);
    my $tempName = "";
    my $lineNr = 0;
    my $tempOs = uc($::configOs);
    my $tempLine;
    while (defined($tempLine = <CONFIGGLOBAL>)) {
        $lineNr++;
        # ignore comment lines (must begin with a #)
        if ($tempLine =~ /^\#/) { next; }
        # ignore empty lines
        if ($tempLine =~ /^\s*$/) { next; }
        # cut all control chars from end of line
        $tempLine = ::chommp($tempLine);
        if ($tempLine =~ /^\[([A-Z0-9_]+)\]\s*$/i) {
            if ($::debug > 3) {
                print("DEBUG:   config line $lineNr:  section '$1'\n");
            }
            if ($tempName ne "") {
                print("\nERROR in global configuration File - line ",
                      $lineNr,":\n");
                print("        Incomplete section for test ",$tempName,"\n");
                close(CONFIGGLOBAL);
                exit 2;
            } else {
                $tempName = $1;
            }
            # [DEFAULT] section has different semantics
            if (uc($tempName) eq "DEFAULT") {
                while (defined($tempLine = <CONFIGGLOBAL>)
                       and $tempLine !~ /^\[\/DEFAULT\]\s*$/i) {
                    $lineNr++;
                    next if ($tempLine =~ /^\s*$/ or $tempLine =~ /^\#/);
                    if ($tempLine =~ /^\s*([A-Z0-9_]+)\s*=\s*(\S.*)$/i) {
                        my $key = uc($1);
                        my $value = ::chommp($2);
                        # remove trailing whitespace from value
                        $value =~ s/^(.+\S)\s+$/$1/;
                        # remove (optional) quotation from value
                        $value =~ s/^\"(.+)\"$/$1/;
                        if ($::debug > 3) {
                            print("DEBUG:   config line $lineNr:",
                                  "  key='$key'  value='$value'\n");
                        }
                        if ($key eq "TEST") {
                            # only set default test if not given on cmdline
                            $::configTestName = $value
                                unless $::configTestName;
                        }
                        elsif ($key eq "ROOT_".uc($::OS_UNIX)) {
                            $::configRoot{$::OS_UNIX} = $value;
                        }
                        elsif ($key eq "ROOT_".uc($::OS_WINDOWS)) {
                            $::configRoot{$::OS_WINDOWS} = $value;
                        }
                        elsif ($key eq "TEMP_".$tempOs) {
                            $::TempRoot = $value unless $::TempRoot;
                        }
                        elsif ($key eq "BIN_".$tempOs) {
                            $::ScriptDir = $value;
                        }
                        elsif ($key eq "ZIPPROG_".$tempOs) {
                            $::ZipProg = $value;
                        }
                        elsif ($key eq "UNZIPPROG_".$tempOs) {
                            $::UnzipProg = $value;
                        }
                        elsif ($key eq "ZIPOPT") {
                            $::ZipOpt = $value;
                        }
                        elsif ($key eq "UNZIPOPT") {
                            $::UnzipOpt = $value;
                        }
                        elsif ($key eq "ZIPDELOPT") {
                            $::ZipDelOpt = $value;
                        }
                        elsif ($key eq "ZIPEXT") {
                            $::ZipExt = $value;
                        }
                        elsif ($key eq "DIRLIST_".$tempOs) {
                            $::DirListDir = $value;
                        }
                        elsif ($key eq "DIR_EVAL") {
                            $::configDirNames{"eval"} = $value;
                        }
                        elsif ($key eq "DIR_LOG") {
                            $::configDirNames{"log"} = $value;
                        }
                        elsif ($key eq "DIR_NEWLOG") {
                            $::configDirNames{"newlog"} = $value;
                        }
                        elsif ($key eq "DIR_TEMPLATE") {
                            $::configDirNames{"template"} = $value;
                        }
                        elsif ($key eq "DIR_KEYS") {
                          $::configDirNames{"keys"} = $value;
                        }
                        elsif ($key eq "DIR_LISTS") {
                            $::configDirNames{"lists"} = $value;
                        }
                        elsif ($key eq "DIR_RESULT") {
                            $::configDirNames{"result"} = $value;
                        }
                        elsif ($key eq "DIR_CSV") {
                            $::configDirNames{"resultCSV"} = $value;
                        }
                        elsif ($key eq "DIR_LOSS_CSV") {
                            $::configDirNames{"resultLossCSV"} = $value;
                        }
                        elsif ($key eq "DIR_REPORT") {
                            $::configDirNames{"report"} = $value;
                        }
                        elsif ($key eq "DIR_INFO") {
                            $::configDirNames{"info"} = $value;
                        }
                        elsif ($key eq "DIR_INFO_VER") {
                            $::configDirNames{"infoVer"} = $value;
                        }
                        elsif ($key eq "DIR_INFO_OPT") {
                            $::configDirNames{"infoOpt"} = $value;
                        }
                        elsif ($key eq "DIR_INFO_PROB") {
                            $::configDirNames{"infoProb"} = $value;
                        }
                        elsif ($key eq "DIR_DOC") {
                            $::configDirNames{"doc"} = $value;
                        }
                        elsif ($key eq "DEBUGFILE") {
                            $::configDebugFile = $value;
                        }
                    }
                }
                # reset tempName to signal end of this section
                $tempName = "";
            }
        } elsif ($tempLine =~ /^($tempOs)\s*=\s*(\S+)\s*$/i) {
            my $tempFile = $2;
            if ($::debug > 3) {
                print("DEBUG:   config line $lineNr:  key='$1'  ",
                      "value='$tempFile'\n");
            }
            $::configAllTests{"$tempName"} = $tempFile;
            $tempName = "";
        }
    }
    # check if last section was complete
    if ($tempName ne "") {
        print("\nERROR in global configuration File - line ",
              $lineNr,":\n");
        print("        Incomplete section for test ",$tempName,"\n");
        close(CONFIGGLOBAL);
        return 0;
    }
    close(CONFIGGLOBAL);

    # get root directory for current system
    $::configRoot = $::configRoot{$::configOs};

    if (not exists($::configDirNames{"eval"})
        or not exists($::configDirNames{"log"})
        or not exists($::configDirNames{"newlog"})
        or not exists($::configDirNames{"info"})
        or not exists($::configDirNames{"doc"})) {
        print("\nERROR in global configuration File:\n");
        print("        missing directory name configuration\n");
        return 0;
    }
    # set root directories
    $::EvalRoot = $::configRoot . $::sep . $::configDirNames{"eval"};
    $::LogRoot = $::configRoot . $::sep . $::configDirNames{"log"};
    $::NewLogRoot = $::configRoot . $::sep . $::configDirNames{"newlog"};
    $::InfoRoot = $::configRoot . $::sep . $::configDirNames{"info"};
    $::DocRoot = $::configRoot . $::sep . $::configDirNames{"eval"} . $::sep . $::configDirNames{"doc"};

    return 1;
}


# read the test specific configuration file
sub readTestConfig {
    # reset flags
    $::loadedTestConfig = 0;

    # set new test locations (directory- and filenames)
    # this is done before calling the new test-specific configuration file
    # to allow it to override these settings
    setTestLocations() or do {
        ::log("ERROR setting new test locations\n\n", 0);
        return 0;
    };

    if (not defined($::configTestFile) or $::configTestFile eq "") {
        ::log("ERROR: test $::configTestName not known\n\n", 0);
        return 0;
    }
    elsif (not -e $::configTestFile) {
        ::log("ERROR: configuration file not found: $::configTestFile\n\n", 0);
        return 0;
    }
    if ($::debug > 2) {
        ::log("parsing configuration file $::configTestFile\n", 0, "DEBUG: ");
    }
    do $::configTestFile or do {
        ::log("ERROR parsing $::configTestFile", 0);
        return 0;
    };

    if ($::configSyntax eq $::OS_WINDOWS) {
        $::rps = $::sep_WINDOWS;
    } else {
        # default is UNIX syntax
        $::rps = $::sep_UNIX;
    }

    $::loadedTestConfig = 1;
    return 1;
}


sub setTestLocations {
    # this is where most processing of the logfiles actually takes place
    # Therefore it should be a local drive no network share !!!
    $::TempBase = $::TempRoot . $::sep . $::configTestName;
    # where to store the evaluation results ?
    $::EvalBase = $::EvalRoot . $::sep . $::configTestName;
    # where are the logfile archives ?
    $::LogBase = $::LogRoot . $::sep . $::configTestName;
    # where do new logfiles appear ?
    # The evaluation system is designed to separate new logfiles from old (already
    # processed) ones to avoid data loss due to human failures
    # Therefore a different directory than LogBase is definitly better
    $::NewLogBase = $::NewLogRoot . $::sep . $::configTestName;
    # where are all the scanner keys ?
    $::KeyDir = $::EvalBase . $::sep . $::configDirNames{"keys"};
    # where are the testbed lists (dirlist, variantlist, sum)?
    $::ListDir = $::EvalBase . $::sep . $::configDirNames{"lists"};
    # where should results be stored ?
    $::ResultDir = $::EvalBase . $::sep . $::configDirNames{"result"};
    # where should results in CSV-Format (for database import) be stored ?
    $::ResultCSVDir = $::EvalBase . $::sep . $::configDirNames{"resultCSV"};
    # where should loss results in CSV-Format (for database import) be stored ?
    $::ResultLossCSVDir = $::EvalBase . $::sep . 
                          $::configDirNames{"resultLossCSV"};
    # where should evaluation results be stored ?
    # All additional evaluation results (problems, productinfo, comparelist, ...)
    # are stored here
    $::ReportDir = $::EvalBase . $::sep . $::configDirNames{"report"};
    # file which lists all testbeds in this test ?
    $::TestTb = $::ListDir . $::sep . "testbed.lst";
    # file which lists all products in this test ?
    $::TestProd = $::ListDir . $::sep . "product.lst";
    # file which lists all operating systems in this test ?
    $::TestOs = $::ListDir . $::sep . "os.lst";
    # file which lists all compression programs used for packed testbeds ?
    $::TestPacked = $::ListDir . $::sep . "packed.lst";
    # file which lists all OS - testbed - product combinations
    # that should NOT be tested
    $::TestNoTest = $::ListDir . $::sep . "no_test.lst";
    # file(s) to write current test status to
    $::TestStatus = $::EvalBase . $::sep . "status.txt";
    $::TestStatusEval = $::EvalBase . $::sep . "status_e.txt";
    # what name should the logfile archives have (without extension) ?
    $::LogArchiveName = "REPORTS";
    # what name should backup archives (used for old logfiles) have (without extension) ?
    $::OldLogArchiveName = "OLD";
    return 1;
}


# get list of all systems/testbeds/products/...
sub getTestLists {
    # reset flags
    $::loadedTestLists = 0;
    $::loadedSum = 0;
    $::loadedPack = 0;
    $::loadedPackCreate = 0;
    $::needsPack = 0;
    # delete old lists
    %::osListAll = ();
    %::osList = ();
    %::listsAll = ();
    %::lists = ();
    %::prodListAll = ();
    %::prodList = ();
    @::noTestListAll = ();
    %::packListAll = ();
    %::packList = ();

    # get new lists
    # The operating system parameter...
    my $ok = getListOs();
    # ...and the testbed parameter (this also sets $::needsPack)...
    if ($ok) { $ok = getListTestbed() }
    # ...and the product parameter...
    if ($ok) { $ok = getListProd() }
    # ...get the 'no test' list...
    if ($ok) { $ok = getListNoTest() }
    # ...get list of compression programs (if required for current test)
    #    These configuration files must exist if the test contains packed testbeds
    if ($ok and $::needsPack) { $ok = getListPack() }
    # ...get archive creation info (if test contains packed testbeds)
    #    It is NO error if the pack-create configuration files are not existent.
    #    Only the 'create' command will then be disabled.
    if ($ok and $::needsPack) { $ok = getListPackCreate() }

    if ($ok) {
        $::loadedTestLists = 1;
    }
    return $ok;
}


# get the lists of all systems and selected systems
sub getListOs {
    open(INITFILE, "< $::TestOs") or do {
        ::log("\nERROR opening $::TestOs: $!\n\n");
        return 0;
    };
    # disable implicit LF <-> CR+LF conversion done on windows systems
    binmode(INITFILE);
    my $tempLine;
    while (defined($tempLine = <INITFILE>)) {
        # ignore comment lines (must begin with a #)
        if ($tempLine =~ /^\#/) { next; }
        # ignore empty lines
        if ($tempLine =~ /^\s*$/) { next; }
        # cut all control chars from end of line
        $tempLine = ::chommp($tempLine);
        # separate all fields (just to check for errors)
        my @tempFields = split(/\s+/, $tempLine);
        # product parameter must be exact 3 chars
        my $tempOs = uc(shift @tempFields);

        if ($tempOs eq "ALL") {
            ::log("\nERROR parsing $::TestOs.\n");
            ::log("'ALL' is not allowed as operating system code.\n");
            close(INITFILE);
            return 0;
        }
        if ($#tempFields < 0 or length($tempOs) != 3) {
            ::log("\nERROR parsing $::TestOs.\n");
            ::log("Expected format for each line is:\n".
                  "3CHAR_CODE systemName\n\n");
            close(INITFILE);
            return 0;
        }
        # this is the product name (may contains spaces)
        my $tempName = join(" ", @tempFields);

        $::osListAll{$tempOs} = $tempName;

        my $selected = 0;
        foreach my $i (@optionOs) {
            if ($i eq $tempOs) {
                $selected = 1;
            }
        }
        if ($optionOs[0] eq "ALL" or $selected) {
            $::osList{$tempOs} = $tempName;
        }
    }
    close(INITFILE);
    # check that we don't operate with zero size lists
    my @tempKeys = keys %::osList;
    if ($#tempKeys < 0) {
        ::log("\nERROR: selected system not found in $::TestOs\n\n");
        return 0;
    }
    @tempKeys = keys %::osListAll;
    if ($#tempKeys < 0) {
        ::log("\nERROR: no content in $::TestOs\n\n");
        return 0;
    }
    if ($::debug > 2) {
        ::log("DEBUG: Systems:");
        foreach my $i (keys %::osList) {
            ::log(" $i");
        }
        ::log($::EOL);
    }

    return 1;
}


# get the lists of all testbeds and selected testbeds
sub getListTestbed {
    # remember missing summary files
    my $noSumFile = 0;
    # get new lists
    open(INITFILE, "< $::TestTb") or do {
        ::log("\nERROR opening $::TestTb: $!\n\n");
        return 0;
    };
    binmode(INITFILE);
    my $tempLine;
    while (defined($tempLine = <INITFILE>)) {
        # ignore comment lines (must begin with a #)
        if ($tempLine =~ /^\#/) { next; }
        # ignore empty lines
        if ($tempLine =~ /^\s*$/) { next; }
        # cut all control chars from end of line
        $tempLine = ::chommp($tempLine);
        # separate all fields (to check for errors)
        my @tempFields = split(/\s+/, $tempLine);
        if ($#tempFields < 5) {
            ::log("\nERROR parsing $::TestTb.\n");
            ::log("Expected format for each line is: ".
                  "ID Type WinPath UnixPath Fullname\n\n");
            close(INITFILE);
            return 0;
        }
        # This is a unique character for this tb (in most cases the driveletter),
        # It is used for some directory names (testers prefer to remember the
        # driveletter instead of the testbed type) and (most important) to
        # name the dirlist files.
        my $tempId = uc(shift @tempFields);
        # This is the type of this testbed given by the syntax defined in
        # 'testbed.txt'. It is used to name the result files, many directory
        # names and internally to identify the testbeds. This is the preferred
        # identifier for testbeds.
        # Its syntax must NOT be changed (but it may be enhanced).
        my $tempType = uc(shift @tempFields);
        # This is the path to the testbed for dos/windows systems
        my $tempPathWin = uc(shift @tempFields);
        # This is the path to the testbed for unix systems (case-sensitive!!!)
        my $tempPathUnix = shift @tempFields;
        # This is the full name describing this testbed (may contains spaces)
        my $tempFullName = join(" ", @tempFields);

        # Build data structures
        $::listsAll{$tempType}->{"path"}->{$::OS_WINDOWS} = $tempPathWin;
        $::listsAll{$tempType}->{"path"}->{$::OS_UNIX} = $tempPathUnix;
        $::listsAll{$tempType}->{"id"} = $tempId;
        $::listsAll{$tempType}->{"name"} = $tempFullName;
        # exclude list is always created, even if there are no excluded samples
        @{$::listsAll{$tempType}->{"excluded"}} = ();
        # check for packed testbeds
        my $tempPackType = substr($tempType, 3, 1);
        if ($tempPackType ne "_") {
            # some extra entries for packed testbeds
            $::needsPack = 1;
            $::listsAll{$tempType}->{"packed"}->{"type"} = $tempPackType;
            my $tempArch = $tempType;
            # packed testbeds  have two evaluation lists:
            # one for the archive contents (this has the normal testbed name)
            # and one for the archives themselves (identified by "_A_" as last
            # 3 chars)
            substr($tempArch, 5, 3) = $::FileExtArchiveFiles;
            $::listsAll{$tempType}->{"packed"}->{"archlist"} = $tempArch;
#            $::listsAll{$tempType}->{"packed"}->{"id"} = "";
            # get summary data for this evaluation list
            my $sumFile = $::ListDir.$::sep.$tempArch.".".$::FileExtSummary;
            # summary data is only required for some commands and shouldn't
            # be loaded for other commands, because it must not exist then
            if (-e $sumFile) {
                open(SUM, "< $sumFile") or do {
                    ::log("ERROR opening $sumFile: $!\n");
                    close(INITFILE);
                    return 0;
                };
                binmode(SUM);
                my ($sumSamples, $sumVir);
                my $inLine = <SUM>;
                (undef, $sumVir, undef, $sumSamples, undef) =
                    split(/\s+/, ::chommp($inLine));
                $::listsAll{$tempType}->{"packed"}->{"sumV"} = $sumVir;
                $::listsAll{$tempType}->{"packed"}->{"sumF"} = $sumSamples;
                close(SUM);
            } else {
                ::log("WARNING: summary file not found: $sumFile\n");
                ::log(" Some commands may be disabled\n", 4);
                $noSumFile = 1;
            }
            # get list of anomalies (if it exists) for this evallist
            my ($anoSample, $anoNr);
            my $anoFile = $::ListDir.$::sep.$tempArch.".".$::FileExtAnomaly;
            # anomaly list is always created, even if there are no anomalies
            # (needed for evaluation)
            %{$::listsAll{$tempType}->{"packed"}->{"ano"}} = ();
            if ( -e $anoFile) {
                open(ANO, "< $anoFile") or do {
                    ::log("ERROR opening $anoFile: $!\n");
                    close(INITFILE);
                    return 0;
                };
                binmode(ANO);
                while (defined($tempLine = <ANO>)) {
                    $tempLine = ::chommp($tempLine);
                    ($anoSample, $anoNr) = split(/\s+/, $tempLine);
                    $anoSample = uc($anoSample);
                    $::listsAll{$tempType}->{"packed"}->{"ano"}->{$anoSample} =
                        $anoNr;
                    if ($::debug > 2) {
                        ::log("added anoSample = '$anoSample'  -  max ".
                              "Samples = $anoNr\n", 0, "DEBUG: ");
                    }
                }
                close(ANO);
            }
            # get list of excluded samples (if it exists) for the archive contents
            # (excluded samples for the archives themselves are appended later)
            my $excSample;
            my $excFile = $::ListDir.$::sep.$tempArch.".".$::FileExtExcluded;
            if ( -e $excFile) {
                open(EXC, "< $excFile") or do {
                    ::log("ERROR opening $excFile: $!\n");
                    close(INITFILE);
                    return 0;
                };
                binmode(EXC);
                while (defined($excSample = <EXC>)) {
                    $excSample = uc(::chommp($excSample));
                    push @{$::listsAll{$tempType}->{"excluded"}}, $excSample;
                    if ($::debug > 2) {
                        ::log("added sample to Excludelist for $tempType: '".
                              "$excSample'\n", 0, "DEBUG: ");
                    }
                }
                close(EXC);
            }
        }
        # get list of excluded samples (if it exists) for this testbed
        my $excSample;
        my $excFile = $::ListDir.$::sep.$tempType.".".$::FileExtExcluded;
        if ( -e $excFile) {
            open(EXC, "< $excFile") or do {
                ::log("ERROR opening $excFile: $!\n");
                close(INITFILE);
                return 0;
            };
            binmode(EXC);
            while (defined($excSample = <EXC>)) {
                $excSample = uc(::chommp($excSample));
                push @{$::listsAll{$tempType}->{"excluded"}}, $excSample;
                if ($::debug > 2) {
                    ::log("added sample to Excludelist for $tempType: ".
                          "'$excSample'\n", 0, "DEBUG: ");
                }
            }
            close(EXC);
        }
        # look for a file containing all evaluation lists for this testbed
        # if it doesn't exists, only one evaluation about the whole tb is done
        my $listslist = $::ListDir.$::sep.$tempType.".".$::FileExtLists;
        if (-e $listslist) {
            # get all evaluation lists for this testbed and create data structure
            open(INITEVAL, "< $listslist") or do {
                ::log("ERROR opening $listslist: $!\n");
                close(INITFILE);
                return 0;
            };
            binmode(INITEVAL);
            my ($tempList, @tempName, $tempId);
            while (defined($tempLine = <INITEVAL>)) {
                $tempLine = ::chommp($tempLine);
                ($tempList, @tempName) = split(/\s+/, $tempLine);
                $tempList = uc($tempList);
                $tempId = substr($tempList, 2, 1);
                # "normal" ID must be empty not underscore
                if ($tempId eq "_") { $tempId = ""; }
                $::listsAll{$tempType}->{"evals"}->{$tempList}->{"name"} =
                        join(" ", @tempName);
                $::listsAll{$tempType}->{"evals"}->{$tempList}->{"id"} = $tempId;
                if ($tempList eq $tempType) {
                    $::listsAll{$tempType}->{"main"} = $tempList;
                } elsif ($tempId eq $::FileIdFP) {
                    $::listsAll{$tempType}->{"fp"} = $tempList;
                }

                # get summary data for this evaluation list
                my $sumFile = $::ListDir.$::sep.$tempList.".".$::FileExtSummary;
                # summary data is only required for some commands and shouldn't
                # be loaded for other commands, because it must not exist then
                if (-e $sumFile) {
                    open(SUM, "< $sumFile") or do {
                        ::log("ERROR opening $sumFile: $!\n");
                        close(INITFILE);
                        close(INITEVAL);
                        return 0;
                    };
                    binmode(SUM);
                    my ($sumSamples, $sumVir);
                    my $inLine = <SUM>;
                    (undef, $sumVir, undef, $sumSamples, undef) =
                        split(/\s+/, ::chommp($inLine));
                    $::listsAll{$tempType}->{"evals"}->{$tempList}->{"sumV"} =
                        $sumVir;
                    $::listsAll{$tempType}->{"evals"}->{$tempList}->{"sumF"} =
                        $sumSamples;
                    close(SUM);
                } else {
                    ::log("WARNING: summary file not found: $sumFile\n");
                    ::log(" Some commands may be disabled\n", 4);
                    $noSumFile = 1;
                }

                # get list of anomalies for this evallist (if it exists)
                my ($anoSample, $anoNr);
                my $anoFile = $::ListDir.$::sep.$tempList.".".$::FileExtAnomaly;
                # anomaly list is always created, even if there are no anomalies
                # (needed for evaluation)
                %{$::listsAll{$tempType}->{"evals"}->{$tempList}->{"ano"}} = ();
                if ( -e $anoFile) {
                    open(ANO, "< $anoFile") or do {
                        ::log("ERROR opening $anoFile: $!\n");
                        close(INITFILE);
                        close(INITEVAL);
                        return 0;
                    };
                    binmode(ANO);
                    while (defined($tempLine = <ANO>)) {
                        $tempLine = ::chommp($tempLine);
                        ($anoSample, $anoNr) = split(/\s+/, $tempLine);
                        $anoSample = uc($anoSample);
                        $::listsAll{$tempType}->{"evals"}->{$tempList}->{"ano"}->{$anoSample} = $anoNr;
                        if ($::debug > 2) {
                            ::log("added anoSample = '$anoSample'  -  max ".
                                  "Samples = $anoNr\n", 0, "DEBUG: ");
                        }
                    }
                    close(ANO);
                }
            }
            close(INITEVAL);
        } else {
            # just define one evaluation list
            my $tempList = $tempType;
            $::listsAll{$tempType}->{"evals"}->{$tempList}->{"name"} =
                $::listsAll{$tempType}->{"name"};
            $::listsAll{$tempType}->{"evals"}->{$tempList}->{"id"} = "";
            $::listsAll{$tempType}->{"main"} = $tempList;
            # set a flag that there's only one evallist
            $::listsAll{$tempType}->{"single"} = $tempList
                unless (exists($::listsAll{$tempType}->{"packed"}));
            # get summary data for this evaluation list
            my $sumFile = $::ListDir.$::sep.$tempList.".".$::FileExtSummary;
            # summary data is only required for some commands and shouldn't
            # be loaded for other commands, cause it must not exist then
            if (-e $sumFile) {
                open(SUM, "< $sumFile") or do {
                    ::log("ERROR opening $sumFile: $!\n");
                    close(INITFILE);
                    return 0;
                };
                binmode(SUM);
                my ($sumSamples, $sumVir);
                my $inLine = <SUM>;
                (undef, $sumVir, undef, $sumSamples, undef) =
                split(/\s+/, ::chommp($inLine));
                $::listsAll{$tempType}->{"evals"}->{$tempList}->{"sumV"} =
                $sumVir;
                $::listsAll{$tempType}->{"evals"}->{$tempList}->{"sumF"} =
                $sumSamples;
                close(SUM);
            } else {
                ::log("WARNING: summary file not found: $sumFile\n");
                ::log(" Some commands may be disabled\n", 4);
                $noSumFile = 1;
            }

            # get list of anomalies for this evallist (if it exists)
            my ($anoSample, $anoNr);
            my $anoFile = $::ListDir.$::sep.$tempList.".".$::FileExtAnomaly;
            # anomaly list is always created, even if there are no anomalies
            # (needed for evaluation)
            %{$::listsAll{$tempType}->{"evals"}->{$tempList}->{"ano"}} = ();
            if ( -e $anoFile) {
                open(ANO, "< $anoFile") or do {
                    ::log("ERROR opening $anoFile: $!\n");
                    close(INITFILE);
                    return 0;
                };
                binmode(ANO);
                while (defined($tempLine = <ANO>)) {
                    $tempLine = ::chommp($tempLine);
                    ($anoSample, $anoNr) = split(/\s+/, $tempLine);
                    $anoSample = uc($anoSample);
                    $::listsAll{$tempType}->{"evals"}->{$tempList}->{"ano"}->{$anoSample} = $anoNr;
                }
                close(ANO);
            }
        }

        my $selected = 0;
        foreach my $i (@optionTb) {
            if ($i eq $tempId or $i eq $tempType) {
                $selected = 1;
            }
        }
        if ($optionTb[0] eq "ALL" or $selected) {
            # don't copy the whole thing - just make a reference
            $::lists{$tempType} = $::listsAll{$tempType};
        }
    }
    close(INITFILE);

    my @tempKeys = keys %::lists;
    if ($#tempKeys < 0) {
        ::log("\nERROR: selected testbed not found in $::TestTb\n\n");
        return 0;
    }
    @tempKeys = keys %::listsAll;
    if ($#tempKeys < 0) {
        ::log("\nERROR: no content in $::TestTb\n\n");
        return 0;
    }
    foreach my $tempList (@tempKeys) {
        my @tempEval = keys %{$::listsAll{$tempList}->{"evals"}};
        if ($#tempEval < 0) {
            ::log("\nERROR: no evaluation list for testbed $tempList\n\n");
            return 0;
        }
    }

    if ($::debug > 2) {
        ::log("Testbeds:\n", 0, "DEBUG: ");
        foreach my $i (keys %::lists) {
            ::log("ID=".$::lists{$i}->{"id"}, 3, "DEBUG: ");
            ::log("  Type=$i", 0);
            ::log(" FullName='".$::lists{$i}->{"name"}."'\n", 0);
            ::log("WinPath='".$::lists{$i}->{"path"}->{$::OS_WINDOWS}."'\n",
                  1, "DEBUG: ");
            ::log(" UnixPath='".$::lists{$i}->{"path"}->{$::OS_UNIX}."'\n",
                  0, "DEBUG: ");
            ::log("Evaluation lists: \n", 4, "DEBUG: ");
            foreach my $j (keys %{$::lists{$i}->{"evals"}}) {
                ::log("$j: id=".$::lists{$i}->{"evals"}->{$j}->{"id"}.
                      " name=".$::lists{$i}->{"evals"}->{$j}->{"name"}."\n",
                      5, "DEBUG: ");
            }
        }
    }

    if ($noSumFile) {
        $::loadedSum = 0;
    } else {
        $::loadedSum = 1;
    }

    return 1;
}


# get the lists of all products and selected products
sub getListProd {
    open(INITFILE, "< $::TestProd") or do {
        ::log("\nERROR opening $::TestProd: $!\n\n");
        return 0;
    };
    binmode(INITFILE);
    my $tempLine;
    while (defined($tempLine = <INITFILE>)) {
        # ignore comment lines (must begin with a #)
        if ($tempLine =~ /^\#/) { next; }
        # ignore empty lines
        if ($tempLine =~ /^\s*$/) { next; }
        # cut all control chars from end of line
        $tempLine = ::chommp($tempLine);
        # separate all fields (just to check for errors)
        my @tempFields = split(/\s+/, $tempLine);
        # product parameter must be exact 3 chars
        my $tempProd = uc(shift @tempFields);

        if ($tempProd eq "ALL") {
            ::log("\nERROR parsing $::TestProd.\n");
            ::log("'ALL' is not allowed as product code.\n");
            close(INITFILE);
            return 0;
        }
        if ($#tempFields < 0 or length($tempProd) != 3) {
            ::log("\nERROR parsing $::TestProd.\n");
            ::log("Expected format for each line is:\n".
                  "3CHAR_CODE fullProductName\n\n");
            close(INITFILE);
            return 0;
        }
        # this is the product name (may contains spaces)
        my $tempName = join(" ", @tempFields);

        $::prodListAll{$tempProd} = $tempName;

        my $selected = 0;
        foreach my $i (@optionProd) {
            if ($i eq $tempProd) {
                $selected = 1;
            }
        }
        if ($optionProd[0] eq "ALL" or $selected) {
            $::prodList{$tempProd} = $tempName;
        }
    }
    close(INITFILE);

    my @tempKeys = keys %::prodList;
    if ($#tempKeys < 0) {
        ::log("\nERROR: selected product not found in $::TestProd\n\n");
        return 0;
    }
    @tempKeys = keys %::prodListAll;
    if ($#tempKeys < 0) {
        ::log("\nERROR: no content in $::TestProd\n\n");
        return 0;
    }
    if ($::debug > 2) {
        ::log("Products:", 0, "DEBUG: ");
        foreach my $i (keys %::prodList) {
            ::log(" $i");
        }
        ::log($::EOL);
    }

    return 1;
}


# get the list of not tested system/testbed/product combinations
sub getListNoTest {
    open(INITFILE, "< $::TestNoTest") or do {
        ::log("\nERROR opening $::TestNoTest: $!\n\n");
        return 0;
    };
    binmode(INITFILE);
    my $tempLine;
    while (defined($tempLine = <INITFILE>)) {
        # ignore comment lines (must begin with a #)
        if ($tempLine =~ /^\#/) { next; }
        # ignore empty lines
        if ($tempLine =~ /^\s*$/) { next; }
        # cut all control chars from end of line
        $tempLine = ::chommp($tempLine);
        # separate all fields (just to check for errors)
        my @tempFields = split(/\s+/, $tempLine);
        # product parameter must be exact 3 chars
        my $tempOs = uc(shift @tempFields);
        my $tempTb = uc(shift @tempFields);
        my $tempProd = uc(shift @tempFields);
        if ($#tempFields >= 0 or length($tempOs) != 3 or length($tempProd) != 3) {
            ::log("\nERROR parsing $::TestNoTest.\n");
            ::log("Expected format for each line is:\n".
                  "OS TESTBED PRODUCT\n\n");
            close(INITFILE);
            return 0;
        }

        my @tempOs =($tempOs);
        my @tempTb =($tempTb);
        my @tempProd = ($tempProd);
        # ATENTION: parameters given as 'ALL' must expand to all possible values
        #           AND an extra entry with 'ALL'
        if ($tempOs eq "ALL") {
            push @tempOs, (keys %::osListAll);
        }
        if ($tempTb eq "ALL") {
            push @tempTb, (keys %::listsAll);
        }
        if ($tempProd eq "ALL") {
            push @tempProd, (keys %::prodListAll);
        }

        foreach my $i (@tempOs) {
            foreach my $j (@tempTb) {
                foreach my $k (@tempProd) {
                    push @::noTestListAll, "$i $j $k";
                }
            }
        }
    }
    close(INITFILE);

    return 1;
}


sub getListPack {
    # get compression programs info
    open(INITPACK, "< $::TestPacked") or do {
        ::log("\nERROR opening $::TestPacked: $!\n\n");
        return 0;
    };
    binmode(INITPACK);
    my $tempLine;
    while (defined($tempLine = <INITPACK>)) {
        # ignore comment lines (must begin with a #)
        if ($tempLine =~ /^\#/) { next; }
        # ignore empty lines
        if ($tempLine =~ /^\s*$/) { next; }
        $tempLine = ::chommp($tempLine);
        # separate all fields (just to check for errors)
        my @tempFields = split(/\s+/, $tempLine);
        if ($#tempFields < 1) {
            ::log("\nERROR parsing $::TestPacked.\n");
            ::log("Expected format for each line is: ".
                  "shortname_of_packer fullname\n".
                  "eg. WRA WinRar\n\n");
            close(INITPACK);
            return 0;
        }
        # this is the 3-char string for this program
        my $tempCode = uc(shift @tempFields);
        if ($tempCode eq "ALL") {
            ::log("\nERROR parsing $::TestPacked.\n");
            ::log("'ALL' is not allowed as packer code.\n");
            close(INITPACK);
            return 0;
        }
        if (length($tempCode) != 3) {
            ::log("\nERROR parsing $::TestPacked.\n");
            ::log("'$tempCode' is not 3 characters long.\n");
            close(INITPACK);
            return 0;
        }

        # this is the program name (may contains spaces)
        my $tempName = join(" ", @tempFields);

        $::packListAll{$tempCode}->{"name"} = $tempName;

        my $selected = 0;
        foreach my $i (@optionPack) {
            if ($i eq $tempCode) {
                $selected = 1;
            }
        }
        if ($optionPack[0] eq "ALL" or $selected) {
            $::packList{$tempCode} = $::packListAll{$tempCode};
        }
    }
    close(INITPACK);

    my @tempKeys = keys %::packList;
    if ($#tempKeys < 0) {
        ::log("\nERROR: no selected compression program found in ".
              "$::TestPacked\n\n");
        return 0;
    }
    @tempKeys = keys %::packListAll;
    if ($#tempKeys < 0) {
        ::log("\nERROR: no content in $::TestPacked\n\n");
        return 0;
    }

    if ($::debug > 2) {
        ::log("Compression programs:\n", 0 , "DEBUG: ");
        foreach my $i (sort keys %::packList) {
            ::log("shortname=$i", 4, "DEBUG: ");
            ::log("name=".$::packList{$i}->{"name"}."\n", 1);
        }
    }

    # Now get archive filenames for all packed testbeds. The filenames may
    # be different for each testbed and are therefore hold together with
    # the testbed data (rather than in the %::packListAll hash)
    foreach my $tb (sort keys %::listsAll) {
        if (not exists($::listsAll{$tb}->{"packed"})) {
            # skip all unpacked testbeds
            next;
        }
        %{$::listsAll{$tb}->{"packed"}->{"prg"}} = ();

        my $TbPacked = $tb;
        substr($TbPacked, 5, 3) = $::FileExtPackedList;
        $TbPacked = $::ListDir . $::sep . $TbPacked . "." . $::FileExtLists;
        open(INITPACKTB, "< $TbPacked") or do {
            ::log("\nERROR opening $TbPacked: $!\n\n");
            return 0;
        };
        binmode(INITPACKTB);
        while (defined($tempLine = <INITPACKTB>)) {
            # ignore comment lines (must begin with a #)
            if ($tempLine =~ /^\#/) { next; }
            # ignore empty lines
            if ($tempLine =~ /^\s*$/) { next; }
            $tempLine = ::chommp($tempLine);
            # separate all fields (just to check for errors)
            my @tempFields = split(/\s+/, $tempLine);
            if ($#tempFields < 1) {
                ::log("\nERROR parsing $TbPacked.\n");
                ::log("Expected format for each line is: ".
                      "shortname_of_packer archive_filename\n".
                      "eg. WRA wrar.rar\n\n");
                close(INITPACKTB);
                return 0;
            }
            # this is the 3-char string for this program as defined
            # in $::TestPacked
            my $tempCode = uc(shift @tempFields);
            if (not exists($::packListAll{$tempCode})) {
                ::log("\nERROR parsing $TbPacked.\n");
                ::log("Unknown code '$tempCode' (not found in ".
                      "$::TestPacked).\n");
                close(INITPACKTB);
                return 0;
            }
            # this is the filename for archives created with this program
            # for this testbed
            my $tempFilename = uc(shift @tempFields);
            $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"filename"} =
                $tempFilename;
            $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"name"} =
                $::packListAll{$tempCode}->{"name"};
        }
        close(INITPACKTB);

        if ($::debug > 2) {
            ::log("Archive filenames in testbed '$tb':\n", 0, "DEBUG: ");
            foreach my $i (keys %{$::listsAll{$tb}->{"packed"}->{"prg"}}) {
                ::log("shortname=$i", 4, "DEBUG: ");
                ::log("filename=".
                      $::listsAll{$tb}->{"packed"}->{"prg"}->{$i}->{"filename"}.
                      "\n", 1);
            }
        }
    }

    $::loadedPack = 1;
    return 1;
}


sub getListPackCreate {
    foreach my $tb (sort keys %::listsAll) {
        if (not exists($::listsAll{$tb}->{"packed"})) {
            # skip all unpacked testbeds
            next;
        }
        # some default values (may be changed by configuration file)
        $::listsAll{$tb}->{"packed"}->{"password"} = "";
        $::listsAll{$tb}->{"packed"}->{"recursive"} = 0;

        # this filename is only used for error messages
        my $tbPacked = $tb;
        substr($tbPacked, 5, 3) = $::FileExtPackedList;
        $tbPacked = $::ListDir . $::sep . $tbPacked . "." . $::FileExtLists;

        my $tbCreate = $::ListDir . $::sep . $tb . "." . $::FileExtPackedList;
        # abort if one of the pack-command files is missing
        if (not -e $tbCreate) {
            $::loadedPackCreate = 0;
            return 1;
        }
        open(CREATEPACK, "< $tbCreate") or do {
            ::log("\nERROR opening $tbCreate: $!\n\n");
            return 0;
        };
        binmode(CREATEPACK);

        if ($::debug > 2) {
            ::log("reading configuration file $tbCreate\n", 0, "DEBUG: ");
        }
        my $tempCode = "";
        my $lineNr = 0;
        my $tempLine;
        LOOPPACK: while (defined($tempLine = <CREATEPACK>)) {
            $lineNr++;
            # ignore comment lines (must begin with a #)
            if ($tempLine =~ /^\#/) { next; }
            # ignore empty lines
            if ($tempLine =~ /^\s*$/) { next; }
            # cut all control chars from end of line
            $tempLine = ::chommp($tempLine);
            if ($tempLine =~ /^\[([A-Z0-9_]+)\]\s*$/i) {
                # found a new section
                if ($::debug > 2) {
                    ::log("config line $lineNr:  section '$1'\n", 1, "DEBUG: ");
                }
                # simple check for required elements in previous section
                if ($tempCode ne "" and $tempCode ne "DEFAULT" and not
                    defined($::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"add"})) {
                    ::log("\nERROR in configuration file $tbCreate - line ".
                          "$lineNr:\n", 0);
                    ::log("Incomplete section for program $tempCode\n", 4);
                    close(CREATEPACK);
                    return 0;
                }

                $tempCode = uc($1);
                # [DEFAULT] section is handled different
                if ($tempCode ne "DEFAULT") {
                    if (not exists($::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode})) {
                        ::log("\nERROR parsing $tbCreate.\n", 0);
                        ::log("Unknown code '$tempCode' (not found in ".
                              "$tbPacked).\n", 3);
                        close(CREATEPACK);
                        return 0;
                    }
                    # default values
                    $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"onlyArch"} = 0;
                    $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"onlyComp"} = 0;
                    $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"onlyRelative"} = 0;
                    $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"wildcard"} = "*.*";
                }
                next LOOPPACK;
            } elsif ($tempLine =~ /^\s*(\S+)\s*=\s*(\S.*)$/i) {
                my $key = uc($1);
                my $value = ::chommp($2);
                $value =~ s/^(.+\S)\s+$/$1/;
                # convert boolean strings to numbers
                if ($value =~ /^false$/i) { $value = 0; }
                elsif ($value =~ /^true$/i) { $value = 1; }
                # remove quotation
                $value =~ s/^\"(.+)\"$/$1/;

                if ($::debug > 2) {
                    ::log("config line $lineNr:  key='$key'  value='".
                          "$value'\n", 1, "DEBUG: ");
                }

                # [DEFAULT] section has different keywords
                if ($tempCode eq "DEFAULT") {
                    if ($key eq "PASSWORD") {
                        $::listsAll{$tb}->{"packed"}->{"password"} = $value;
                        next LOOPPACK;
                    }
                    elsif ($key eq "RECURSIVE") {
                        if ($value > 0) {
                            $::listsAll{$tb}->{"packed"}->{"recursive"} =
                                $value;
                            next LOOPPACK;
                        } else {
                            ::log("ERROR: Illegal value for 'Recursive'".
                                  " for $tempCode (line $lineNr)\n");
                        }
                    }
                } else {
                    # all other sections
                    if ($key eq "ADD") {
                        $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"add"} = $value;
                        # check for file placeholders in commandline
                        my $testVar = $value;
                        my $failed = 0;
                        if (not $testVar =~ s/\%i//) { $failed = 1; }
                        if (not $testVar =~ s/\%o//) { $failed = 1; }
                        if ($::listsAll{$tb}->{"packed"}->{"type"} eq "E"
                            and not $testVar =~ s/\%p//) {
                            $failed = 1;
                        }
                        if ($testVar =~ /\%(i|o|p)/) { $failed = 2; }
                        if ($failed == 1) {
                            ::log("ERROR: Missing file placeholders in add command".
                                  " for $tempCode (line $lineNr)\n", 0);
                            ::log("You must specify the position of the ".
                                  "output file with %o and the input file(s)".
                                  " with %i\n", 3);
                        } elsif ($failed == 2 and $::debug) {
                            ::log("WARNING: multiple identical file ".
                                  "placeholders in add command for $tempCode".
                                  " (line $lineNr)\n");
                        }
                        next LOOPPACK;
                    }
                    elsif ($key eq "ONLYARCH") {
                        $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"onlyArch"} = $value;
                        next LOOPPACK;
                    }
                    elsif ($key eq "ONLYCOMP") {
                        $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"onlyComp"} = $value;
                        next LOOPPACK;
                    }
                    elsif ($key eq "ONLYRELATIVE") {
                        $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"onlyRelative"} = $value;
                        next LOOPPACK;
                    }
                    elsif ($key eq "WILDCARD") {
                        $::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"wildcard"} = $value;
                        next LOOPPACK;
                    }
                }
            }
            # abort in all other cases
            ::log("\nERROR in configuration file $tbCreate - line $lineNr:".
                  "\n");
            ::log("Can't interpret line '$tempLine'\n", 3);
            close(CREATEPACK);
            return 0;
        }

        # check if last section was complete
        if (not defined($::listsAll{$tb}->{"packed"}->{"prg"}->{$tempCode}->{"add"})) {
            ::log("\nERROR in configuration file $tbCreate - line $lineNr:".
                  "\n");
            ::log("Incomplete section for program $tempCode\n", 3);
            close(CREATEPACK);
            return 0;
        }
        close(CREATEPACK);
    }

    $::loadedPackCreate = 1;
    return 1;
}


sub defineFilenameParts {
    # list-filename extensions
    $::FileExtText = "TXT";
    $::FileExtLog = "LOG";
    $::FileExtLists = "LST";
    $::FileExtVariants = "VAR";
    $::FileExtSamples = "SAM";
    $::FileExtSummary = "SUM";
    $::FileExtFP = "FP";
    $::FileExtAnomaly = "ANO";
    $::FileExtExcluded = "EXC";
    $::FilePreDirlist = "DIR_";

    # filename extensions for scanner key API
    $::FileExtReport = "FUL";
    $::FileExtNormUnsorted = "NRU";
    $::FileExtScannedOK = "SBO";
    $::FileExtAuswert = "AUS";
    $::FileExtAuswertRedundant = "AUR";
    $::FileExtReportRem = "RES";
    $::FileExtControl = "CON";
    $::FileExtCutList = "CUT";
    $::FileExtCorrupt = "COR";

    # filenames for testreport parts
    $::FileReportProblems = "8PROBLMS.TXT";
    $::FileReportProdInfo = "A2SCANLS.TXT";
    $::FileCmpAll = "RES_COMP.TXT";
    # readme.txt file to include in exported logfile archives
    $::FileExportLogReadme = "log_read.txt";
    $::FileExportLogPreReadme = "log_pre.txt";

    # other filename parts
    $::FileExtStatus = "_STAT.TXT";
    $::FileExtCmpF = "CPF";
    $::FileExtCmpV = "CPV";
    $::FileExtCmpWinF = "WCF";
    $::FileExtCmpWinV = "WCV";
    $::FileExtScan1 = "S1";
    $::FileExtScan2 = "N1";
    $::FileExtScan3 = "N2";
    $::FileExtPSL = "TXT";
    $::FileExtNorm = "NRM";
    $::FileExtNormRemPacked = "RPN";
    $::FileExtScanOKRemPacked = "RPS";
    $::FileExtPackedAdd = "ADD";
    $::FileExtPackedOKAdd = "ADO";
    $::FileExtDetFiles = "D_F";
    $::FileExtMissedFiles = "MIS";
    $::FileExtDetFPFiles = "DFP";
    $::FileExtMissedFPFiles = "MFP";
    $::FileExtAnoFiles = "A_F";
    $::FileExtGarbageFiles = "G_F";
    $::FileExtExcludedFiles = "E_F";
    $::FileExtScanFiles = "S_F";
    $::FileExtOKFiles = "O_F";
    $::FileExtInfAndOK = "INO";
    $::FileExtPostscan = "PSL";
    $::FileExtGarbage = "S_G";
    $::FileExtEval = "EV1";
    $::FileExtRem = "RE1";
    $::FileExtDetVir = "DET";
    $::FileExtUnrId = "U_I";
    $::FileExtMultInf = "M_I";
    $::FileExtMultScan = "M_S";
    $::FileExtFilteredAno = "ANO";
    $::FileExtResult = "DAT";
    $::FileExtArchiveFiles = "_A_";
    $::FileExtPackedList = "PAK";
    $::FileIdFP = "F";
}


sub defineStatusCodes {
    # format: $::statusCode{"code"} = "description"
    %::statusCodes = (
       "-S1" => "waiting for scan report",
       "S1" => "evaluate scan report",
       "S1E" => "quality control for scan report",
       "OK S1E" => "scan report okay without postscans",
       "-S2" => "waiting for 1. postscan report",
       "S2" => "evaluate 1. postscan report",
       "S2E" => "quality control for 1. postscan report",
       "OK S2E" => "scan report okay with 1 postscan",
       "-S3" => "waiting for 2. postscan report",
       "S3" => "evaluate 2. postscan report",
       "S3E" => "quality control for 2. postscan report",
       "OK S3E" => "scan report okay with 2 postscans",
       "NO TEST" => "not tested",
       "NO REPORT" => "product failed to create a report",
    );
}


sub defineCommands {
    # format:
    # $::cmds{"cmd"}->{"desc"} = STRING
    #               ->{"exec"}  = \&doCommand;
    #               ->{"opt"}   = \$var_with_options;
    #               ->{"type"}  = STRING (?)
    #               ->{"label"} = STRING
    # optional:
    #               ->{"notGUI"} => "true"
    #               ->{"needsNoGlobal"} => "true"
    #               ->{"needsNoConfig"} => "true"
    #               ->{"needsSum"} => "true"
    #               ->{"needsPackCreate"} => "true"
    #               ->{"ask"} => "true"
    #               ->{"admin"} => "true"
    #
    # type is a list of characters denotig the accepted parameters:
    #  "o" = operating system
    #  "t" = testbed
    #  "p" = product
    #  "a" = packer
    #  "s" = status
    %::cmds = (
        "help" => {
                 "exec" => \&::help,
                 "label" => "Help",
                 "desc" => "Show help",
                 "type" => "",
                 "needsNoGlobal" => 1,
                 },
        "new" => {
                 "exec" => \&::makeNew,
                 "label" => "New",
                 "desc" => "Set up configuration for new test",
                 "type" => "",
                 "needsNoConfig" => 1,
                 "admin" => 1,
                 },
        "update" => {
                 "exec" => \&::makeUpdate,
                 "label" => "Update",
                 "desc" => "Update (or create) testbed lists",
                 "type" => "t",
                 "ask" => 1,
                 "admin" => 1,
                 },
        "archive" => {
                 "exec" => \&::archive,
                 "label" => "Archive",
                 "desc" => "Move scan reports to archive but don't evaluate them yet",
                 "type" => "otp",
                 },
        "eval" => {
                 "exec" => \&::evaluate,
                 "label" => "Eval",
                 "desc" => "Evaluate new scan reports (not evaluated before)",
                 "type" => "otp",
                 "needsSum" => 1,
                 },
        "reeval" => {
                 "exec" => \&::reeval,
                 "label" => "Re-Eval",
                 "desc" => "Evaluate scan reports that have been evaluated before",
                 "type" => "otp",
                 "needsSum" => 1,
                 },
        "set ok" => {
                 "exec" => \&::setOk,
                 "label" => "Set OK",
                 "desc" => "Set status of a report to 'OK' (report passed QA)",
                 "type" => "otp",
                 "ask" => 1,
                 },
        "set postscan" => {
                 "exec" => \&::setPostscan,
                 "label" => "Set Postscan",
                 "desc" => "Set status of a report to 'postscan'",
                 "type" => "otp",
                 "ask" => 1,
                 },
        "reset" => {
                 "exec" => \&::changeStatus,
                 "opt" => \$::resetStatus,
                 "label" => "Reset",
                 "desc" => "Reset status of a product to a given status (if possible)",
                 "type" => "sotp",
                 "ask" => 1,
                 "admin" => 1,
                 },
        "reset ok" => {
                 "exec" => \&::resetOk,
                 "label" => "Reset OK",
                 "desc" => "Reset status of a finished logfile to its previous status",
                 "type" => "otp",
                 "ask" => 1,
                 },
        "reset postscan" => {
                 "exec" => \&::resetPostscan,
                 "label" => "Reset Postscan",
                 "desc" => "as 'reset ok' but for logfile with waiting postscan",
                 "type" => "otp",
                 "ask" => 1,
                 },
        "result" => {
                 "exec" => \&::reportResult,
                 "label" => "Result",
                 "desc" => "Write result tables and do status report",
                 "type" => "ot",
                 "needsSum" => 1,
                 },
        "resultcsv" => {
                 "exec" => \&::exportResult,
                 "label" => "Result CSV",
                 "desc" => "Write CSV result tables for database import",
                 "type" => "ot",
                 "needsSum" => 1,
                 },
        "status" => {
                 "exec" => \&::reportStatus,
                 "label" => "Status",
                 "desc" => "Write a summary what's still missing for the current test",
                 "type" => "",
                 },
        "extract" => {
                 "exec" => \&::extract,
                 "label" => "Extract",
                 "desc" => "Extract eval data from server to tempDir on the client",
                 "type" => "otp",
                 },
        "show config" => {
                 "exec" => \&::showConfig,
                 "label" => "Show Config",
                 "desc" => "Display configuration settings for current test",
                 "type" => "",
                 },
        "show tests" => {
                 "exec" => \&::showTests,
                 "label" => "Show Tests",
                 "desc" => "Display list of all known tests",
                 "type" => "",
                 "needsNoConfig" => 1,
                 },
        "show systems" => {
                 "exec" => \&::showSystems,
                 "label" => "Show Systems",
                 "desc" => "Display list of all operating systems for current test",
                 "type" => "",
                 },
        "show testbeds" => {
                 "exec" => \&::showTestbeds,
                 "label" => "Show Testbeds",
                 "desc" => "Display list of all testbeds for current test",
                 "type" => "",
                 },
        "show products" => {
                 "exec" => \&::showProducts,
                 "label" => "Show Products",
                 "desc" => "Display list of all products for current test",
                 "type" => "",
                 },
        "show pack" => {
                 "exec" => \&::showPacked,
                 "label" => "Show Pack",
                 "desc" => "Display list of all compression programs for current test",
                 "type" => "",
                 },
        "show packcmd" => {
                 "exec" => \&::showPackedCmd,
                 "label" => "Show Packcommands",
                 "desc" => "Display details of all compression programs",
                 "type" => "",
                 "needsPackCreate" => 1,
                 },
        "show dump" => {
                 "exec" => \&::dumpData,
                 "label" => "Show Dump",
                 "desc" => "Display internal data structures used for current test",
                 "type" => "otpa",
                 "needsSum" => 1,
                 },
        "show status" => {
                 "exec" => \&::showStatus,
                 "label" => "Show Status",
                 "desc" => "Display current status for selected test",
                 "type" => "",
                 },
        "report rank" => {
                 "exec" => \&::reportRank,
                 "label" => "Report Rank",
                 "desc" => "Create ranking lists (only for heureka tests)",
                 "type" => "",
                 "needsSum" => 1,
                 },
        "report info" => {
                 "exec" => \&::reportInfo,
                 "label" => "Report Info",
                 "desc" => "Create report file with product information",
                 "type" => "",
                 },
        "report problems" => {
                 "exec" => \&::reportProblems,
                 "label" => "Report Problems",
                 "desc" => "Create report file with all test problems",
                 "type" => "",
                 },
        "report logs" => {
                 "exec" => \&::exportLogs,
                 "label" => "Report Logs",
                 "desc" => "Export logfiles for distribution to product vendors",
                 "type" => "",
                 },
        "report missed" => {
                 "exec" => \&::exportMissed,
                 "label" => "Report Missed",
                 "desc" => "Export missed samples for distribution to product vendors",
                 "type" => "",
                 },
        "report compare" => {
                 "exec" => \&::reportCompare,
                 "label" => "Report Compare",
                 "desc" => "Create compare tables for results under different systems",
                 "type" => "",
                 "needsSum" => 1,
                 },
        "check compare" => {
                 "exec" => \&::checkCompare,
                 "label" => "Check compare",
                 "desc" => "Create file with all result data (for QA)",
                 "type" => "",
                 "needsSum" => 1,
                 },
        "clean" => {
                 "exec" => \&::cleanTest,
                 "label" => "Clean",
                 "desc" => "Delete all redundant evaluation files",
                 "type" => "",
                 "ask" => 1,
                 "admin" => 1,
                 },
        "cleanall" => {
                 "exec" => \&::cleanTestAll,
                 "label" => "Clean All",
                 "desc" => "Delete all files from this test except result data !!!",
                 "type" => "",
                 "ask" => 1,
                 "admin" => 1,
                 },
        "create" => {
                 "exec" => \&::createPacked,
                 "label" => "Create",
                 "desc" => "Create packed testbeds",
                 "type" => "ta",
                 "needsPackCreate" => 1,
                 "ask" => 1,
                 "admin" => 1,
                 },
        "gui" => {
                 "exec" => \&::startGUI,
                 "label" => "GUI",
                 "desc" => "Start graphical user interface",
                 "type" => "",
                 "needsNoConfig" => 1,
                 "notGUI" => 1,
                 },
    );
}


sub defineHelp {
    $::usage[0] =
    "Usage: $::prog [options] COMMAND [OS] [TB] [PROD] [PACK]\n" .
    " COMMAND   the command that should be executed is one of:\n";
    foreach my $cmd (sort keys %::cmds) {
            $::usage[0] .= sprintf("   %-15s - %s\n",
                                 $cmd, $::cmds{$cmd}->{"desc"});
    }
    $::usage[0] .= " Note: Some commands are only available in ".
                   "'Administrator Mode'\n";

    $::usage[1] =
    " The 'show' command needs an additional first parameter: The " .
               "configuration\n" .
    "           data to show. This is one of:\n" .
    "           config, dump, tests, systems, testbeds, products, pack, " .
               "packcmd\n" .
    " The 'reset' command needs an additional first parameter that specifies " .
               "the new\n" .
    "           status. Whitespace in status strings must be omitted (eg. " .
               "NOTEST)!\n".
    "           If no STATUS parameter is given, '-S1' is the default.\n" .
    " OS        Restrict evaluation to some operating systems (eg. " .
               "'DOS,LNX').\n" .
    " TB        Restrict evaluation to some testbeds.\n" .
    "           The testbed code can either be the drive letter (eg. H), \n" .
    "           the type (eg. mi______) or the dirname (eg. macr_itw).\n" .
    " PROD      Restrict evaluation to some products (eg. 'FPR')\n" .
    " PACK      Restrict evaluation to some compression programs (only for " .
                "'create')\n" .
    "\n" .
    "All OS, TB and PROD parameters are case insensitive and must\n".
    "be entered in the specified order.\n" .
    "Each parameter can be a single code, a comma-separated list of codes\n".
    "or 'all' to select all systems, testbeds or products.\n" .
    'Note: On windows systems, you must surround the list with "", '.
    'eg. "DOS,LNX".' . "\n" .
    "Valid codes can be listed with the 'show' commands.\n" .
    "If the last parameters are omitted, they are taken as 'ALL':\n" .
    "  '$::prog eval' is the same as '$::prog eval all all all' and\n" .
    "  '$::prog eval dos macr' is equal to '$::prog eval dos macr all'.\n";


    $::usage[2] =
    "Options:\n" .
    " -n  -test=NAME     work on test 'NAME' (NAME is case sensitive!)\n" .
    " -c  -config=FILE   use alternative global configuration file\n" .
    " -testconfig=FILE   use alternative test specific configuration file\n" .
    " -t  -temp=DIR      use directory 'DIR' for temporary files\n" .
    " -q  -quiet          suppress all additional output\n" .
    " -d  -debug         debug mode (be more verbose)\n" .
    " -p  -pause         pause after every screen of output (only for help ".
                        "output)\n" .
    " -s  -skipReload    don't get logfiles from archive, re-use files from ".
                        "the\n" .
    "                    temporary dirctory (this is only valid for 'eval' " .
                        "and \n" .
    "                    'reeval' commands)\n" .
    "\nOther options (exit immediatly):\n" .
    " -v  -version       version information \n" .
    " -h                 short help\n" .
    " -help              what you see is what you get :-)\n";


    $::usage[3] =
    "To avoid confusion how to use all these commands and options,\n".
    "here is a list of all possible combinations:\n" .
    "\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] help\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-p] help\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show config\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show dump\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show tests\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show systems\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show testbeds\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show products\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show pack\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] show packcmd\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] new\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] update\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] update [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] create\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] create [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] create [TB] [PACK]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] clean\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] cleanall\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] archive [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] archive OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] archive OS TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-s] eval [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-s] eval OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-s] eval OS TB [PROD]\n";


    $::usage[4] =
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-s] reeval [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-s] reeval OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] [-s] reeval OS TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] extract [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] extract OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] extract OS TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] set ok OS TB PROD\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] set postscan OS TB PROD\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset [STATUS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset STATUS [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset STATUS OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset STATUS OS TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset ok [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset ok OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset ok OS TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset postscan [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset postscan OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] reset postscan OS TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] result [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] result OS [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] status [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] report rank [OS]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] report info\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] report problems\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] report compare [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] report compare TB [PROD]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] check compare [TB]\n" .
    "$::prog [-d] [-c FILE] [-n name] [-t dir] check compare TB [PROD]\n";
}


sub displayUsage {
    my $short = 0;
    if ($_[0] eq "short") { $short = 1; }

    ::log("$::prog $::version\n\n");
    ::log($::usage[0]);
    if ($short) {
        ::log("Type '$::prog help' for more information\n");
    } else {
        for (my $i = 1; $i < 5; $i++) {
            if ($::pause) {
                print("\n --- press ENTER for next page ---");
                getc;
                print("\n");
            }
            ::log($::usage[$i]);
       }
    }
    ::log("\n");
}


sub help {
    displayUsage("");
    return 0;
}

